

def _fullname(o):
	module = o.__class__.__module__
	if module is None or module == str.__class__.__module__:
		return o.__class__.__name__		# Avoid reporting __builtin__
	else:
		#items = module.split(".")
		#if items[-1] == o.__class__.__name__:
		#	return module
		#else:
		#	return module + '.' + o.__class__.__name__
		return module + '.' + o.__class__.__name__
#




def getTypeAsStr(value, fqn:bool = False, bAllowNull:bool = True):
	assert isinstance(fqn, bool)

	if value is None:
		if bAllowNull:
			return "(null)"
		else:
			raise Exception("Value is (null)!")

	if isinstance(value, bool):
		return "bool"

	if isinstance(value, int):
		return "int"

	if isinstance(value, float):
		return "float"

	if isinstance(value, str):
		return "str"

	if isinstance(value, (list, tuple)):
		assert len(value) > 0

		elementType = None
		firstValue = None
		for v in value:
			if elementType is None:
				if v is not None:
					elementType = type(v)
					firstValue = v
			else:
				if v is not None:
					assert elementType == type(v)

		if firstValue is None:
			raise Exception("Only (null)s!")

		return getTypeAsStr(v, fqn, False) + "[]"

	if isinstance(value, dict):
		assert len(value) > 0
		keyType = None
		elementType = None
		firstKey = None
		firstValue = None
		for k in value.keys():
			v = value[k]

			if keyType is None:
				assert k is not None
				keyType = type(k)
				firstKey = k
			else:
				assert keyType == type(k)

			if elementType is None:
				if v is not None:
					elementType = type(v)
					firstValue = v
			else:
				if v is not None:
					assert elementType == type(v)

		if firstKey is None:
			raise Exception("Only (null)s!")
		if firstValue is None:
			raise Exception("Only (null)s!")

		return "dict<" + getTypeAsStr(firstKey, fqn, False) + "," + getTypeAsStr(firstValue, fqn, False) + ">"

	if fqn:
		s = _fullname(value)
	else:
		s = value.__class__.__name__
		pos = s.rfind(".")
		if pos:
			s = s[pos + 1:]
	return s
#










import os
import atexit



def _delPIDFile(filePath:str):
	if os.path.isfile(filePath):
		os.unlink(filePath)
#


def writeProcessPIDFile(filePath:str):
	dirPath = os.path.dirname(filePath)
	os.makedirs(dirPath, exist_ok=True)

	pid = os.getpid()
	with open(filePath, "w") as f:
		f.write(str(pid) + "\n")

	atexit.register(_delPIDFile, filePath)
#







import signal



class DelayedKeyboardInterrupt(object):

	def __init__(self, log, message = "SIGINT received. Delaying keyboard interrupt."):
		self.__bSignalRaised = False
		self.__signal_received = False
		self.__log = log
		self.__message = message
		self.__old_handler = None

	def __enter__(self):
		self.__bSignalRaised = False
		self.__signal_received = False
		self.__old_handler = signal.getsignal(signal.SIGINT)
		signal.signal(signal.SIGINT, self.handler)

	def handler(self, sig, frame):
		self.__signal_received = (sig, frame)
		if self.__message != None:
			self.__log.debug(self.__message)

	def __exit__(self, type, value, traceback):
		signal.signal(signal.SIGINT, self.__old_handler)
		if self.__signal_received:
			self.__old_handler(*self.__signal_received)



















import os
import sys
import socket
import struct
import select
import time
import json




if sys.platform == "win32":
	# On Windows, the best timer is time.clock()
	__dDefaultTimer = time.clock
else:
	# On most other platforms the best timer is time.time()
	__dDefaultTimer = time.time

# From /usr/include/linux/icmp.h; your milage may vary.
__ICMP_ECHO_REQUEST = 8 # Seems to be the same on Solaris.




def __checksum(source_string):
	"""
	I'm not too confident that this is right but testing seems
	to suggest that it gives the same answers as in_cksum in ping.c
	"""
	sum = 0
	countTo = (len(source_string)/2)*2
	count = 0
	while count<countTo:
		thisVal = source_string[count + 1]*256 + source_string[count]
		sum = sum + thisVal
		sum = sum & 0xffffffff # Necessary?
		count = count + 2

	if countTo<len(source_string):
		sum = sum + source_string[len(source_string) - 1]
		sum = sum & 0xffffffff # Necessary?

	sum = (sum >> 16)  +  (sum & 0xffff)
	sum = sum + (sum >> 16)
	answer = ~sum
	answer = answer & 0xffff

	# Swap bytes. Bugger me if I know why.
	answer = answer >> 8 | (answer << 8 & 0xff00)

	return answer
#



#
# Receives a single ping.
#
# @param		socket my_socket		The socket to receive the ping from
# @param		int ID					The ID the ping packet must have (otherwise it will be rejected)
# @param		int timeout				The number of seconds to wait for a ping received
# @return		tuple<str,int>			Returns either <c>(None, None)</c> or a tuple containing:
#										* the sender IP address as string
#										* the delay in seconds
#
def __receive_one_ping(my_socket, ID, timeout):
	timeLeft = timeout
	while True:
		startedSelect = __dDefaultTimer()
		whatReady = select.select([my_socket], [], [], timeLeft)
		howLongInSelect = (__dDefaultTimer() - startedSelect)
		if whatReady[0] == []: # Timeout
			return (None, None)

		timeReceived = __dDefaultTimer()
		recPacket, addr = my_socket.recvfrom(1024)
		#print("-- received from: " + addr[0])
		icmpHeader = recPacket[20:28]
		type, code, checksum, packetID, sequence = struct.unpack(
			"bbHHh", icmpHeader
		)
		# Filters out the echo request itself.
		# This can be tested by pinging 127.0.0.1
		# You'll see your own request
		if type != 8 and packetID == ID:
			bytesInDouble = struct.calcsize("d")
			timeSent = struct.unpack("d", recPacket[28:28 + bytesInDouble])[0]
			return (addr[0], timeReceived - timeSent)

		timeLeft = timeLeft - howLongInSelect
		if timeLeft <= 0:
			return (None, None)
#



#
# Receives a set of pings.
#
# @param		socket my_socket		The socket to receive the ping from
# @param		int ID					The ID the ping packet must have (otherwise it will be rejected)
# @param		int timeout				The number of seconds to wait for a ping received
# @param		int count				The number of ping packets expected. This routine will return if this number of
#										ping packets has successfully been received.
# @return		dict					Returns a dictionary containing:
#										* keys: the IP addresses as strings
#										* values: the delay of the ping packet
#
def __receive_multiple_pings(my_socket, ID, timeout, count):
	"""
	receive a number of pings ping from the socket.
	"""
	ret = {}
	timeLeft = timeout
	while True:
		startedSelect = __dDefaultTimer()
		whatReady = select.select([my_socket], [], [], timeLeft)
		howLongInSelect = (__dDefaultTimer() - startedSelect)
		if whatReady[0] == []: # Timeout
			return ret

		timeReceived = __dDefaultTimer()
		recPacket, addr = my_socket.recvfrom(1024)
		#print("-- received from: " + addr[0])
		icmpHeader = recPacket[20:28]
		type, code, checksum, packetID, sequence = struct.unpack(
			"bbHHh", icmpHeader
		)
		# Filters out the echo request itself.
		# This can be tested by pinging 127.0.0.1
		# You'll see your own request
		if type != 8 and packetID == ID:
			bytesInDouble = struct.calcsize("d")
			timeSent = struct.unpack("d", recPacket[28:28 + bytesInDouble])[0]
			ret[addr[0]] = timeReceived - timeSent
			count -= 1
			if count == 0:
				return ret

		timeLeft = timeLeft - howLongInSelect
		if timeLeft <= 0:
			return ret
#



#
# Send a single ping packet.
#
# @param		socket my_socket		The socket to send the packet on.
# @param		str dest_ip_addr		A IP adddress (!) to send the ping packet to.
# @param		int ID					An identifier to send with the ping message
# @return		bool					Returns <c>True</c> on success, <c>False</c> on error.
#										Such an error occurred if sending a ping failed f.e. because
#										there is no network connection available.
#
def __send_one_ping(my_socket, dest_ip_addr, ID):
	# Header is type (8), code (8), checksum (16), id (16), sequence (16)
	my_checksum = 0

	# Make a dummy heder with a 0 checksum.
	header = struct.pack("bbHHh", __ICMP_ECHO_REQUEST, 0, my_checksum, ID, 1)
	bytesInDouble = struct.calcsize("d")
	data = bytes((192 - bytesInDouble) * "Q", 'utf-8')
	data = struct.pack("d", __dDefaultTimer()) + data

	# Calculate the checksum on the data and the dummy header.
	my_checksum = __checksum(header + data)

	# Now that we have the right checksum, we put that in. It's just easier
	# to make up a new header than to stuff it into the dummy.
	header = struct.pack(
		"bbHHh", __ICMP_ECHO_REQUEST, 0, socket.htons(my_checksum), ID, 1
	)
	packet = header + data
	try:
		my_socket.sendto(packet, (dest_ip_addr, 1)) # Don't know about the 1
		return True
	except OSError:
		return False
#



#
# Perform a single ping to the specified address.
# You must be root to invoke this function.
#
# @param		string destinationAddress	The destination host to send the ping packet to. This can either be a host name or an IP address.
# @param		int timeout					The timeout in seconds to wait for a response.
# @return		tuple<int,str>				Returns a tuple of two values:
#											* the sender IP address as string
#											* the delay in seconds (possibly <c>None</c> if the destination did not reply)
#
def pingSingleHost(destinationAddress, timeout):
	if os.geteuid() != 0:
		raise Exception("Pings can only be sent as root!")

	icmp = socket.getprotobyname("icmp")
	try:
		my_socket = socket.socket(socket.AF_INET, socket.SOCK_RAW, icmp)
	except PermissionError as e:
		#e.args = (e.args if e.args else tuple()) + ((
		#	" - Note that ICMP messages can only be sent from processes"
		#	" running as root."
		#),)
		raise

	my_ID = os.getpid() & 0xFFFF

	destIPAddr = socket.gethostbyname(destinationAddress)
	__send_one_ping(my_socket, destIPAddr, my_ID)
	senderIPAddr, delay = __receive_one_ping(my_socket, my_ID, timeout)

	my_socket.close()
	return destIPAddr, delay
#


#
# Send a ping to multiple hosts at the same time.
# You must be root to invoke this function.
#
# @param		string[] destinationAddresses	An iterable of destination addresses. These can either be a host name or an IP address.
#												You should not attempt to use more than 300 IP addresses in this list as it has been found
#												to otherwise overload the internal operating system buffers for the network interface.
#												(This limit roughly applies to regular Linux systems. Please note that this limit might be
#												a bit different.)
# @param		int timeout						The timeout in seconds to wait for the responses.
# @param		bool orderByIPAddr				This routine performs address resolution. This parameter decides about the grouping of the results returned.
#												If <c>True</c> is specified here this routine returns a map containing IP addresses as
#												keys and tuples of host names and delays as values.
#												If <c>False</c> is specified here this routine returns a map containing the host names
#												as keys and tuples of IP addresses and delays as values.
# @return		dict							Returns a dictionary with tuples. Depending on the parameter <c>orderByIPAddr</c> the keys
#												are either host names or IP addresses, the values tuples of host names or IP addresses as first element
#												and the ping delays in seconds as the second element. This second element is <c>None</c> if the
#												destination host did not reply.
#
def pingMultipleHosts(destinationAddresses, timeout:int, orderByIPAddr:bool = False):
	if os.geteuid() != 0:
		raise Exception("Pings can only be sent as root!")

	icmp = socket.getprotobyname("icmp")
	try:
		my_socket = socket.socket(socket.AF_INET, socket.SOCK_RAW, icmp)
	except PermissionError as e:
		#e.args = (e.args if e.args else tuple()) + ((
		#	" - Note that ICMP messages can only be sent from processes"
		#	" running as root."
		#),)
		raise

	my_ID = os.getpid() & 0xFFFF

	orgAddresses = []
	ipAddresses = []
	dataMap = {}
	for dest_addr in destinationAddresses:
		orgAddresses.append(dest_addr)
		ipAddress = socket.gethostbyname(dest_addr)
		ipAddresses.append(ipAddress)
		#print(dest_addr + " :: " + destIPAddr)
		dataMap[ipAddress] = (dest_addr, None)

	#n = 0
	nPingsSend = 0
	nPingsSendSucceeded = 0
	for destIPAddr in ipAddresses:
		#print(n)
		#n += 1
		#print("-- " + str(my_socket.getsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF)))
		nPingsSend += 1
		if __send_one_ping(my_socket, destIPAddr, my_ID):
			nPingsSendSucceeded += 1
	if nPingsSendSucceeded == 0:
		# we did not succeed in sending a single ping, probably we're not connected right now.
		receiverMap = {}
	else:
		# at least one ping could be sent.
		receiverMap = __receive_multiple_pings(my_socket, my_ID, timeout, len(orgAddresses))
		#print(receiverMap)

	for ipAddr in receiverMap.keys():
		orgAddr = dataMap[ipAddr][0] if ipAddr in dataMap else None
		dataMap[ipAddr] = (orgAddr, receiverMap[ipAddr])

	if not orderByIPAddr:
		# reorder
		dataMap2 = {}
		for ipAddr in dataMap:
			orgAddr, delay = dataMap[ipAddr]
			dataMap2[orgAddr] = (ipAddr, delay)
		dataMap = dataMap2

	my_socket.close()
	return dataMap
#












import re
import collections

from .Stack import Stack
from .TypedValue import TypedValue







#
# This defines a token data structure. This structure has the following fields:
#
# @attribute		str type		A type identifier that specifies the type of the token
# @attribute		obj value		The value of the token (can be of any data type)
# @attribute		int lineNo		The line number (counted from 1)
# @attribute		int colNo		The character number within the line (counted from 1)
#
Token = collections.namedtuple('Token', ['type', 'value', 'lineNo', 'colNo'])






#
# RegEx based tokenizer: This tokenizer uses regular expressions in order to create tokens.
#
class RegExBasedTokenizer(object):

	#
	# Initialization method
	#
	# @param		str[] patternDefs		A list of pattern definitions. Each pattern definition should be a tuple or list of one of the following structures:
	#										* 2 items
	#											* (required) the token type to use on match
	#											* (required) the pattern to match
	#										* 4 items
	#											* (required) the token type to use on match
	#											* (optional) if not None a pattern to match but that will not be part of the token content
	#											* (required) the pattern to match (which will be the content of the token)
	#											* (optional) if not None a pattern to match but that will not be part of the token content
	#
	def __init__(self, patternDefs):
		assert isinstance(patternDefs, (tuple, list))
		assert len(patternDefs) > 0

		p = ""
		for patternDef in patternDefs:
			assert isinstance(patternDef, (tuple, list))
			if len(p) > 0:
				p += "|"
			if len(patternDef) == 2:
				p += '(?P<' + patternDef[0] + '>' + patternDef[1] + ')'
			elif len(patternDef) == 4:
				if patternDef[1] != None:
					p += '(?:' + patternDef[1] + ')'
				p += '(?P<' + patternDef[0] + '>' + patternDef[2] + ')'
				if patternDef[3] != None:
					p += '(?:' + patternDef[3] + ')'
			else:
				raise Exception("Invalid pattern definition: " + str(patternDef))

		self.__rawPatterns = p
		self.__compiledPatterns = None
		self.__typeParsingDelegates = {}
	#



	@property
	def isCompiled(self):
		return self.__compiledPatterns != None
	#



	def compile(self):
		rawPatterns = self.__rawPatterns
		rawPatterns += "|(?P<NEWLINE>\n)"
		rawPatterns += "|(?P<SPACE>[\t ]+)"
		rawPatterns += "|(?P<ERROR>.)"
		try:
			self.__compiledPatterns = re.compile(rawPatterns)
		except Exception as e:
			print("==== COMPILE ERROR")
			print("==== " + e.msg)
			print("==== " + rawPatterns)
			s = ""
			for i in range(0, e.colno):
				s += " "
			s += "^"
			print(s)
			raise
	#



	def addTokenPattern(self, tokenType:str, patternRegExStr):
		assert isinstance(tokenType, str)

		bErr = False
		if isinstance(patternRegExStr, (tuple, list)):
			if len(patternRegExStr) == 1:
				if isinstance(patternRegExStr[0], str):
					patternRegExStr = patternRegExStr[0]
				else:
					bErr = True
			elif len(patternRegExStr) == 3:
				pass
			else:
				bErr = True
		elif isinstance(patternRegExStr, str):
			pass
		else:
			bErr = True
		if bErr:
			raise Exception("Invalid patternRegExStr specified!")

		if len(self.__rawPatterns) > 0:
			self.__rawPatterns += "|"
		if isinstance(patternRegExStr, str):
			self.__rawPatterns += '(?P<' + tokenType + '>' + patternRegExStr + ')'
		else:
			if patternRegExStr[0] != None:
				self.__rawPatterns += '(?:' + patternRegExStr[0] + ')'
			self.__rawPatterns += '(?P<' + tokenType + '>' + patternRegExStr[1] + ')'
			if patternRegExStr[2] != None:
				self.__rawPatterns += '(?:' + patternRegExStr[2] + ')'

		self.__compiledPatterns = None
	#



	def registerTypeParsingDelegate(self, tokenTypeMain, tokenTypeExtra, delegate):
		assert isinstance(tokenTypeMain, str)
		assert isinstance(tokenTypeExtra, (type(None), str))
		assert callable(delegate)

		if tokenTypeExtra is None:
			tokenType = tokenTypeMain
		else:
			tokenType = tokenTypeMain + "_" + tokenTypeExtra
		self.__typeParsingDelegates[tokenType] = delegate
	#



	#
	# @return		Token[] tokens			Returns token objects according to the pattern defined during initialization.
	#
	def tokenize(self, text, bEmitWhiteSpaces = False, bEmitNewLines = False):
		if not self.isCompiled:
			self.compile()

		lineNo = 1
		line_start = 0
		for mo in self.__compiledPatterns.finditer(text):
			tokenType = mo.lastgroup
			value = mo.group(tokenType)
			columnNo = mo.start() - line_start + 1
			if tokenType == 'NEWLINE':
				if bEmitNewLines:
					yield Token(tokenType, value, lineNo, columnNo)
				line_start = mo.end()
				lineNo += 1
			elif tokenType == 'SPACE':
				if bEmitWhiteSpaces:
					yield Token(tokenType, value, lineNo, columnNo)
			elif tokenType == 'ERROR':
				raise RuntimeError("Tokenization error encountered at " + str(lineNo) + ":" + str(columnNo) + "!")
			else:
				d = self.__typeParsingDelegates.get(tokenType, None)
				if d != None:
					value = d(value)
				pos = tokenType.find("_")
				# tokenTypeExtra = None
				#if pos >= 0:
					#tokenTypeExtra = tokenType[pos + 1:]
					#tokenType = tokenType[0:pos]
				# s = "" if tokenTypeExtra == None else tokenTypeExtra
				# print("---- " + tokenTypeMain + ":" + s + " ---- \"" + tokenText + "\"")
				if pos >= 0:
					tokenType = tokenType[0:pos]
				yield Token(tokenType, value, lineNo, columnNo)
	#



	#
	# Override this method to perform correct value parsing. This method is called for every
	#
	def _parseValue(self, tokenTypeMain, tokenTypeExtra, tokenText):
		return tokenText
	#

#






#
# Abstract base class for all token pattern classes.
#
class AbstractTokenPattern(object):

	def __init__(self):
		self.__tags = None
	#

	#
	# Tries to match at the specified position.
	#
	# @param		list tokens				The list of tokens.
	# @param		int offset				Where to try a match.
	# @param		dict defaults			A set of defaults the return data map should be populated with.
	# @return		tuple					Returns a tuple containing the following data:
	#										* bool ---- Indicates success or failure
	#										* int ---- The number of tokens eaten, <c>0</c> if nothing could be eaten
	#										* dict ---- A dictionary containing the parsed data or <c>None</c> if nothing could be eaten
	#
	def tryMatch(self, tokens, offset = 0, defaults = None):
		assert isinstance(tokens, list)
		assert isinstance(offset, int)
		assert isinstance(defaults, (type(None), dict))

		stack = Stack()
		(bResult, n) = self._tryMatch(tokens, offset, stack)
		#print("---- tryMatch: " + str(bResult) + ", " + str(n))
		if bResult:
			ret = {
				"lineNo": tokens[offset].lineNo,
				"colNo": tokens[offset].colNo,
			}
			if defaults != None:
				ret.update(defaults)

			for (entryType, a, b, c) in stack:

				if (entryType == "v") or (entryType == "vv"):

					# there is a value to store.

					# step 1: get the data to store
					token, varName, bVarIsArray = a, b, c
					if entryType == "v":
						value = token.value
					else:
						value = TypedValue(token.type, token.value)

					# step 2: actually store the value
					v = ret.get(varName, None)
					if v is None:
						if bVarIsArray:
							ret[varName] = [ value ]
						else:
							ret[varName] = value
					else:
						if isinstance(v, list):
							v.append(value)
							ret[varName] = v
						else:
							ret[varName] = [ v, value ]

				elif entryType == "t":
					a, varName, value = a, b, c
					ret[varName] = value

				else:
					raise Exception("Implementation Error!")

			return (True, n, ret)
		else:
			return (False, 0, None)
	#

	@property
	def tags(self):
		if self.__tags != None:
			return dict(self.__tags)
		else:
			return None
	#

	#
	# Associate a key with a value at this pattern. If parsing succeeds the result data map will receive this key-value-pair.
	#
	def setTag(self, tagName, tagValue):
		if self.__tags is None:
			self.__tags = {}
		self.__tags[tagName] = tagValue
		return self
	#

	def _addTagsToStack(self, stack):
		if self.__tags != None:
			for tagName in self.__tags.keys():
				stack.push(("t", None, tagName, self.__tags[tagName]))
	#

	def _tryMatch(self, tokens, pos, stack):
		pass
	#

	#
	# Derive a pattern from this pattern with specific variations.
	#
	# @param	str assignToVar				If set the result data map will receive the value directly as it is.
	# @param	str assignToVarTyped		If set the result data map will receive the value wrapped in an instance of <c>TypedValue</c>.
	# @param	bool bVarIsArray			Normally values in the result data map will be wrapped into an array only if multiple values occur with the
	#										same key. If this argument is set to <c>True</c> an array is created by default, even if it will receive only one
	#										value. This is useful if the values parsed will represent array data anyway.
	#
	def derive(self, assignToVar = None, assignToVarTyped = None, bVarIsArray = False):
		pass
	#

#






#
# An atom-like token pattern.
#
class TokenPattern(AbstractTokenPattern):

	#
	# @param		str tokenType			(required) The type identifier of a token, e.g. "int"
	# @param		str tokenValue			(optional) The text value the token is expected to have for a match.
	# @param		str assignToVar			(optional) On match use the value as the string it is within the token.
	# @param		str assignToVarTyped	(optional) On match use the value but wrap the token text into `TypedValue`,
	#										using the token type as type information for `TypedValue'. Thus this way type information
	#										is preserved.
	# @param		bool bVarIsArray		(optional) This variable is an array. If it is an array, data is stacked as an array.
	#
	def __init__(self, tokenType:str, tokenValue:str = None, assignToVar:str = None, assignToVarTyped:str = None, bVarIsArray:bool = False):
		super().__init__()

		assert isinstance(tokenType, str)
		assert isinstance(tokenValue, (type(None), str))
		assert isinstance(assignToVar, (type(None), str))
		assert isinstance(assignToVarTyped, (type(None), str))
		assert isinstance(bVarIsArray, bool)

		self.__tokenType = tokenType
		self.__tokenValue = tokenValue
		self.__assignToVar = assignToVar
		self.__assignToVarTyped = assignToVarTyped
		self.__bVarIsArray = bVarIsArray
	#

	def _tryMatch(self, tokens:list, pos:int, stack:Stack):
		assert isinstance(tokens, list)
		assert isinstance(pos, int)
		assert isinstance(stack, Stack)

		if pos >= len(tokens):
			return (False, 0)
		if tokens[pos].type != self.__tokenType:
			return (False, 0)
		if (self.__tokenValue != None) and (tokens[pos].value != self.__tokenValue):
			return (False, 0)
		if self.__assignToVar != None:
			stack.push(("v", tokens[pos], self.__assignToVar, self.__bVarIsArray))
		if self.__assignToVarTyped != None:
			stack.push(("vv", tokens[pos], self.__assignToVarTyped, self.__bVarIsArray))
		self._addTagsToStack(stack)
		return (True, 1)
	#

	#
	# Derive a new token pattern from this object, modifying some attributes in the process.
	#
	def derive(self, assignToVar:str = None, assignToVarTyped:str = None, bVarIsArray:bool = False):
		assert isinstance(assignToVar, (type(None), str))
		assert isinstance(assignToVarTyped, (type(None), str))
		assert isinstance(bVarIsArray, bool)

		return TokenPattern(self.__tokenType, self.__tokenValue, assignToVar, assignToVarTyped, bVarIsArray)
	#

#






#
# A token sequence pattern. The complete sequence must match, regardless of what kind of element it consists.
#
class TokenPatternSequence(AbstractTokenPattern):

	#
	# @param		AbstractTokenPattern[] tokenPatterns			(required) A sequence of token patterns to match against.
	#
	def __init__(self, tokenPatterns:list):
		super().__init__()

		assert isinstance(tokenPatterns, list)
		for t in tokenPatterns:
			assert isinstance(t, AbstractTokenPattern)
		self.__tokenPatterns = tokenPatterns
		self.__len = len(tokenPatterns)
	#

	def _tryMatch(self, tokens, pos, stack):
		assert isinstance(tokens, list)
		assert isinstance(pos, int)
		assert isinstance(stack, Stack)

		m = stack.mark()
		orgPos = pos
		for i in range(0, self.__len):
			(bResult, n) = self.__tokenPatterns[i]._tryMatch(tokens, pos, stack)
			assert isinstance(bResult, bool)
			assert isinstance(n, int)

			if not bResult:
				m.resetToMark()
				return (False, 0)
			pos += n
		if pos - orgPos > 0:
			self._addTagsToStack(stack)
			return (True, pos - orgPos)
		else:
			return (False, 0)
	#

	#
	# Derive a new token pattern from this object, modifying some attributes in the process.
	#
	def derive(self, assignToVar:str = None, assignToVarTyped:str = None, bVarIsArray:bool = False):
		assert isinstance(assignToVar, (type(None), str))
		assert isinstance(assignToVarTyped, (type(None), str))
		assert isinstance(bVarIsArray, bool)

		tokenPatterns = []
		for tokenPattern in self.__tokenPatterns:
			tokenPatterns.append(tokenPattern.derive(assignToVar, assignToVarTyped, bVarIsArray))
		return TokenPatternSequence(tokenPatterns)
	#

#






#
# A token pattern repeat pattern. The specified patteren must match at least once.
#
# @param		AbstractTokenPattern tokenPattern			(required) A token pattern that might be repeated multiple times.
#
class TokenPatternRepeat(AbstractTokenPattern):

	#
	# @param		AbstractTokenPattern tokenPattern			(required) A token pattern to match one or more times.
	#
	def __init__(self, tokenPattern:AbstractTokenPattern):
		super().__init__()

		assert isinstance(tokenPattern, AbstractTokenPattern)
		self.__tokenPattern = tokenPattern
	#

	def _tryMatch(self, tokens, pos, stack):
		assert isinstance(tokens, list)
		assert isinstance(pos, int)
		assert isinstance(stack, Stack)

		orgPos = pos
		nCountLoops = 0
		while True:
			(bResult, n) = self.__tokenPattern._tryMatch(tokens, pos, stack)
			assert isinstance(bResult, bool)
			assert isinstance(n, int)

			if not bResult:
				break
			pos += n
			nCountLoops += 1
		if nCountLoops > 0:
			self._addTagsToStack(stack)
			return (True, pos - orgPos)
		else:
			return (False, 0)
	#

	#
	# Derive a new token pattern from this object, modifying some attributes in the process.
	#
	def derive(self, assignToVar:str = None, assignToVarTyped:str = None, bVarIsArray:bool = False):
		assert isinstance(assignToVar, (type(None), str))
		assert isinstance(assignToVarTyped, (type(None), str))
		assert isinstance(bVarIsArray, bool)

		return TokenPatternRepeat(self.__tokenPattern.derive(assignToVar, assignToVarTyped, bVarIsArray))
	#

#






#
# A token pattern alternative pattern. One of the specified patterns must match.
#
class TokenPatternAlternatives(AbstractTokenPattern):

	#
	# @param		AbstractTokenPattern[] tokenPatterns			(required) A set of token patterns to match against.
	#
	def __init__(self, tokenPatterns:list):
		super().__init__()

		assert isinstance(tokenPatterns, list)
		for t in tokenPatterns:
			assert isinstance(t, AbstractTokenPattern)
		self.__tokenPatterns = tokenPatterns
		self.__len = len(tokenPatterns)
	#

	def _tryMatch(self, tokens, pos, stack):
		assert isinstance(tokens, list)
		assert isinstance(pos, int)
		assert isinstance(stack, Stack)

		m = stack.mark()
		for i in range(0, self.__len):
			(bResult, n) = self.__tokenPatterns[i]._tryMatch(tokens, pos, stack)
			assert isinstance(bResult, bool)
			assert isinstance(n, int)

			if bResult:
				self._addTagsToStack(stack)
				return (True, n)
			m.resetToMark()
		return (False, 0)
	#

	def derive(self, assignToVar = None, assignToVarTyped = None, bVarIsArray = False):
		assert isinstance(assignToVar, (type(None), str))
		assert isinstance(assignToVarTyped, (type(None), str))
		assert isinstance(bVarIsArray, bool)

		tokenPatterns = []
		for tokenPattern in self.__tokenPatterns:
			tokenPatterns.append(tokenPattern.derive(assignToVar, assignToVarTyped, bVarIsArray))
		return TokenPatternAlternatives(tokenPatterns)
	#

#






#
# A token pattern optional pattern. The specified patteren must match either zero times or once.
#
class TokenPatternOptional(AbstractTokenPattern):

	#
	# @param		AbstractTokenPattern tokenPattern			(required) A token patterns to match against.
	#
	def __init__(self, tokenPattern):
		super().__init__()

		assert isinstance(tokenPattern, AbstractTokenPattern)
		self.__tokenPattern = tokenPattern
	#

	def _tryMatch(self, tokens, pos, stack):
		assert isinstance(tokens, list)
		assert isinstance(pos, int)
		assert isinstance(stack, Stack)

		(bResult, n) = self.__tokenPattern._tryMatch(tokens, pos, stack)
		assert isinstance(bResult, bool)
		assert isinstance(n, int)

		self._addTagsToStack(stack)
		return (True, n)
	#

	def derive(self, assignToVar = None, assignToVarTyped = None, bVarIsArray = False):
		assert isinstance(assignToVar, (type(None), str))
		assert isinstance(assignToVarTyped, (type(None), str))
		assert isinstance(bVarIsArray, bool)

		return TokenPatternOptional(self.__tokenPattern.derive(assignToVar, assignToVarTyped, bVarIsArray))
	#

#












import os
import stat
import collections










FileSystemStats = collections.namedtuple("FileSystemStats", [
	"bytesTotal", "bytesUsed", "bytesFreeSystem", "bytesFreeUser", "rateBytesUsed",
	"inodesTotal", "inodesUsed", "inodesFree", "rateInodesUsed"
])

def getFileSystemStats(mountPointPath):
	statvfs = os.statvfs(mountPointPath)

	bytesTotal = statvfs.f_frsize * statvfs.f_blocks		# Size of filesystem in bytes
	bytesFreeSystem = statvfs.f_frsize * statvfs.f_bfree	# Actual number of free bytes
	bytesFreeUser = statvfs.f_frsize * statvfs.f_bavail		# Number of free bytes that ordinary users are allowed to use (excl. reserved space)
	bytesUsed = bytesTotal - bytesFreeSystem
	rateBytesUsed = ((bytesTotal - bytesFreeUser) / bytesTotal) if bytesTotal > 0 else 1

	inodesTotal = statvfs.f_files    					    # inodes
	inodesFree = statvfs.f_ffree	   						# free inodes
	inodesUsed = inodesTotal - inodesFree
	rateInodesUsed = (inodesUsed / inodesTotal) if inodesTotal > 0 else 1

	return FileSystemStats(
		bytesTotal, bytesUsed, bytesFreeSystem, bytesFreeUser, rateBytesUsed,
		inodesTotal, inodesUsed, inodesFree, rateInodesUsed
	)
#

def findMountPoint(path:str):
	path = os.path.abspath(path)
	orig_dev = os.stat(path).st_dev

	while path != '/':
		pdir = os.path.dirname(path)
		if os.stat(pdir).st_dev != orig_dev:
			# we crossed the device border
			break
		path = pdir
	return path
#



def isDirEmpty(path:str):
	path = os.path.abspath(path)
	if os.path.isdir(path):
		entries = os.listdir(path)
		if entries:
			return False
		else:
			return True
	else:
		raise Exception("Not a directory: " + path)
#



#
# Recursively get the size of the specified folder in bytes.
#
# @param		str dirPath		(required) The directory to recursively get the size from
# @param		str mode		(optional) Either "exact" or "block" (default) to either sum up the exact file sizes
#								or to sum up the size the files occupy on disk.
# @return		int totalSize	Returns the total size of all files.
#
def __old_getFolderSize(dirPath:str, mode:str = "block") -> int:
	assert isinstance(dirPath, str)
	assert os.path.isdir(dirPath)

	assert isinstance(mode, str)
	assert mode in [ "exact", "block" ]

	def _getFolderSize_exact(dirPath:str) -> int:
		nBytes = os.path.getsize(dirPath)
		for item in os.listdir(dirPath):
			itempath = os.path.join(dirPath, item)
			if os.path.isfile(itempath):
				nBytes += os.path.getsize(itempath)
			elif os.path.isdir(itempath):
				nBytes += _getFolderSize_exact(itempath)
		return nBytes
	#

	def _getFolderSize_block(dirPath:str, blockSize:int) -> int:
		nBlocks = (os.path.getsize(dirPath) + blockSize - 1) // blockSize
		for item in os.listdir(dirPath):
			itempath = os.path.join(dirPath, item)
			if os.path.isfile(itempath):
				nBlocks += (os.path.getsize(itempath) + blockSize - 1) // blockSize
			elif os.path.isdir(itempath):
				nBlocks += _getFolderSize_block(itempath, blockSize)
		return nBlocks
	#

	if mode == "exact":
		return _getFolderSize_exact(dirPath)
	else:
		blockSize = os.stat(dirPath).st_blksize
		nBytes = (os.path.getsize(dirPath) + blockSize - 1) // blockSize
		return _getFolderSize_block(dirPath, blockSize) * blockSize
#



#
# Recursively get the size of the specified folder in bytes.
#
# @param		str dirPath		(required) The directory to recursively get the size from
# @param		str mode		(optional) Either "exact" or "block" (default) to either sum up the exact file sizes
#								or to sum up the size the files occupy on disk.
# @return		int totalSize	Returns the total size of all files.
#
def getFolderSize(dirPath:str, mode:str = "block") -> int:
	assert isinstance(dirPath, str)
	assert os.path.isdir(dirPath)

	assert isinstance(mode, str)
	assert mode in [ "exact", "block" ]

	def _getFolderSize_exact(dirPath:str) -> int:
		nBytes = 0
		for fe in os.scandir(dirPath):
			_stat = fe.stat(follow_symlinks = False)
			nBytes += _stat.st_size
			if stat.S_ISDIR(_stat.st_mode):
				nBytes += _getFolderSize_exact(fe.path)
		return nBytes
	#

	def _getFolderSize_block(dirPath:str, blockSize:int) -> int:
		nBlocks = 0
		for fe in os.scandir(dirPath):
			_stat = fe.stat(follow_symlinks = False)
			nBlocks += ((_stat.st_size + blockSize - 1) // blockSize)
			if stat.S_ISDIR(_stat.st_mode):
				nBlocks += _getFolderSize_block(fe.path, blockSize)
		return nBlocks
	#

	if mode == "exact":
		nBytes = os.path.getsize(dirPath)
		return nBytes + _getFolderSize_exact(dirPath)
	else:
		blockSize = os.stat(dirPath).st_blksize
		nBytes = ((os.path.getsize(dirPath) + blockSize - 1) // blockSize) * blockSize
		return nBytes + _getFolderSize_block(dirPath, blockSize) * blockSize
#









def getFileNameWithoutExtension(fileName:str) -> str:
	fileName = os.path.basename(fileName)
	pos = fileName.rfind(".")
	if pos < 0:
		return fileName
	else:
		return fileName[:pos]
#


#
# Get the file name extension.
#
def getFileNameExtension(fileName:str, bIncludeDot:bool = True) -> str:
	fileName = os.path.basename(fileName)
	pos = fileName.rfind(".")
	if pos < 0:
		return None
	else:
		if bIncludeDot:
			return fileName[pos:]
		else:
			return fileName[pos + 1:]
#


def _recursiveWalk0(baseDirPath:str, currentDirPath:str, subDirParts:tuple, dirFilter, fileFilter):
	# print(">>", currentDirPath)
	entries = sorted(list(os.scandir(currentDirPath)), key=lambda x: x.name)
	for fe in entries:
		# yield baseDirPath, subDirParts, fe
		# print(fe)
		if fe.is_dir():
			# print("DIR:", fe)
			bDescend = (dirFilter is None) or dirFilter(fe)
			if bDescend:
				yield from _recursiveWalk0(baseDirPath, fe.path, subDirParts + ( fe.name, ), dirFilter, fileFilter)
		elif fe.is_file():
			# print("FILE:", fe)
			bAccept = (fileFilter is None) or fileFilter(fe)
			if bAccept:
				yield baseDirPath, subDirParts, fe
		else:
			pass
#

#
# Generator that recursively returns files.
#
# @param	str dirPath				The directory to scan recursively
# @param	callable dirFilter		A delegate that receives a directory entry object and returns `True` or `False` depending wether to further descend into that directory
# @param	callable fileFilter		A delegate that receives a file entry object and returns `True` or `False` depending wether to return this entry or not
# @return	str baseDirPath			The directory that is the root of the recursive descend
# @return	str[] subDirParts		The name of the subdirectories
# @return	io.FileEntry fe			A file entry data structure as returned by `os.scan()`
#
def recursiveWalk(dirPath:str, dirFilter, fileFilter):
	assert isinstance(dirPath, str)
	assert os.path.isdir(dirPath)
	if dirFilter is not None:
		assert callable(dirFilter)
	if fileFilter is not None:
		assert callable(fileFilter)

	yield from _recursiveWalk0(dirPath, dirPath, tuple(), dirFilter, fileFilter)
#















__version__ = "0.2022.2.6"



import os
_bIsPOSIX = os.name == "posix"



from .PIDFile import writeProcessPIDFile
from . import file_rw
from . import array
from . import oop
from . import check
if _bIsPOSIX:
	from . import mac
	from . import ip
from . import ping
from . import re
from . import hex
from . import reflection
from . import pathutils
if _bIsPOSIX:
	from . import fsutils
from . import rng
from . import datatypes
from . import python
from .GracefullyHandleKeyboardInterrupt import GracefullyHandleKeyboardInterrupt
from .DelayedKeyboardInterrupt import DelayedKeyboardInterrupt
from .GracefullyHandleInterrupts import GracefullyHandleInterrupts
from .dtutils import dtutils
if _bIsPOSIX:
	from .arp import arp, cachedarp, ArpRecord
from .Cache import Cache
from .ChangedFlag import ChangedFlag
from .Stack import Stack
from .TypedValue import TypedValue
from .EnumBase import EnumBase
from . import tokenizer
from .MutableString import MutableString
from .TextCanvas import TextCanvas
from .TextTable import TextTable, TextTableCell
from .CmdLineParser import CmdLineParser
from .StateManager import StateManager
from .ObservableEvent import ObservableEvent
from .Timer import Timer
from .PersistentProperties import PersistentProperties
from .AsyncRunner import AsyncRunner
from .TimeLimitedCache import TimeLimitedCache
from .RoundRobinSequence import RoundRobinSequence
from .MultiCounterDict import MultiCounterDict
from .pythonmodules import PythonModuleInfo, PythonModules
from .ChModValue import ChModValue
from . import users
from .showCapacityProgress import *
from .DataMatrix import DataMatrix
from . import processes
from .RandomStateID import RandomStateID
from .TextOutputBuffer import TextOutputBuffer
from .VolatileValue import VolatileValue
from .Bytes import Bytes
from .AmountOfBytes import AmountOfBytes
from .ImplementationError import ImplementationError
from .TerminationFlag import TerminationFlag
from .InterruptedException import InterruptedException
from .deprecated import deprecated
from .TimeStamp import TimeStamp
from . import duration
from . import deferred



# the data node implementation has been moved to an own package: jk_datanodes
#from .datanodes import EnumNodeType, DataNode, DataNodeDef


import re
import stat
import typing



#
# This class represents a mode value such as in <c>stat.st_mode</c>.
# It provides easy ways for parsing and modification.
#
class ChModValue:

	################################################################
	## Constants
	################################################################

	################################################################
	## Constructor
	################################################################

	def __init__(self,
			v = None,
			userR:bool = False,
			userW:bool = False,
			userX:bool = False,
			groupR:bool = False,
			groupW:bool = False,
			groupX:bool = False,
			otherR:bool = False,
			otherW:bool = False,
			otherX:bool = False,
		):

		if v is None:
			self.userR = userR
			self.userW = userW
			self.userX = userX
			self.groupR = groupR
			self.groupW = groupW
			self.groupX = groupX
			self.otherR = otherR
			self.otherW = otherW
			self.otherX = otherX
		elif isinstance(v, str):
			self.fromStr(v)
		elif isinstance(v, int):
			self.fromInt(v)
		else:
			raise Exception("Unexected type: " + repr(type(v)))
	#

	################################################################
	## Properties
	################################################################

	################################################################
	## Helper Methods
	################################################################

	def __setBit(self, value, bitNo):
		mask = 1 << bitNo		# Compute mask, an integer with just bit 'index' set.
		value |= mask			# If x was True, set the bit indicated by the mask.
		return value
	#

	################################################################
	## Public Methods
	################################################################

	def cloneObject(self):
		return ChModValue(self.toInt())
	#

	def __str__(self):
		v = [ "-" ] * 9
		if self.userR:
			v[0] = "r"
		if self.userW:
			v[1] = "w"
		if self.userX:
			v[2] = "x"
		if self.groupR:
			v[3] = "r"
		if self.groupW:
			v[4] = "w"
		if self.groupX:
			v[5] = "x"
		if self.otherR:
			v[6] = "r"
		if self.otherW:
			v[7] = "w"
		if self.otherX:
			v[8] = "x"
		return "".join(v)
	#

	def __repr__(self):
		return self.__str__()
	#

	def __int__(self):
		return self.toInt()
	#

	def __ne__(self, other) -> bool:
		if isinstance(other, int):
			return self.toInt() != (other & 0o777)
		elif isinstance(other, str):
			return self.__ne__(ChModValue(other))
		elif isinstance(other, ChModValue):
			return (self.userR != other.userR) \
				or (self.userW != other.userW) \
				or (self.userX != other.userX) \
				or (self.groupR != other.groupR) \
				or (self.groupW != other.groupW) \
				or (self.groupX != other.groupX) \
				or (self.otherR != other.otherR) \
				or (self.otherW != other.otherW) \
				or (self.otherX != other.otherX)
		else:
			return True
	#

	def __eq__(self, other) -> bool:
		if isinstance(other, int):
			return self.toInt() == (other & 0o777)
		elif isinstance(other, str):
			return self.__eq__(ChModValue(other))
		elif isinstance(other, ChModValue):
			return (self.userR == other.userR) \
				and (self.userW == other.userW) \
				and (self.userX == other.userX) \
				and (self.groupR == other.groupR) \
				and (self.groupW == other.groupW) \
				and (self.groupX == other.groupX) \
				and (self.otherR == other.otherR) \
				and (self.otherW == other.otherW) \
				and (self.otherX == other.otherX)
		else:
			return False
	#

	def toStr(self):
		return self.__str__()
	#

	def toStrChMod(self) -> str:
		vU = ""
		if self.userR:
			vU += "r"
		if self.userW:
			vU += "w"
		if self.userX:
			vU += "x"
		vG = ""
		if self.groupR:
			vG += "r"
		if self.groupW:
			vG += "w"
		if self.groupX:
			vG += "x"
		vO = ""
		if self.otherR:
			vO += "r"
		if self.otherW:
			vO += "w"
		if self.otherX:
			vO += "x"
		return "u=" + vU + ",g=" + vG + ",o=" + vO
	#

	#
	# Accept data from the specified integer value.
	#
	def fromInt(self, v:int):
		assert isinstance(v, int)
		if v < 0:
			raise Exception("Unrecognized int!")

		v = v & 0o777

		self.userR = v & stat.S_IRUSR == stat.S_IRUSR
		self.userW = v & stat.S_IWUSR == stat.S_IWUSR
		self.userX = v & stat.S_IXUSR == stat.S_IXUSR
		self.groupR = v & stat.S_IRGRP == stat.S_IRGRP
		self.groupW = v & stat.S_IWGRP == stat.S_IWGRP
		self.groupX = v & stat.S_IXGRP == stat.S_IXGRP
		self.otherR = v & stat.S_IROTH == stat.S_IROTH
		self.otherW = v & stat.S_IWOTH == stat.S_IWOTH
		self.otherX = v & stat.S_IXOTH == stat.S_IXOTH

		return self
	#

	#
	# Accept data from the specified string value.
	# The data in this object is set to exactly match the specified argument.
	# An exception is thrown if the specified string could not be parsed.
	#
	# @param	str s			A string to parse. You should specify something like this:
	#							* "-rwxr--r--" (10 characters, the first must be '-')
	# 							* "rwxr--r--" (9 characters)
	#							* an octal number such as "755"
	#							* "a=rw", "ug=rwx", "o=x", etc.
	#
	def fromStr(self, v:str):
		assert isinstance(v, str)
		v = v.lower()

		if re.match("^[0-7][0-7][0-7]$", v):
			# ocal form
			return self.fromInt(int(v, 8))

		if re.match("^[r-][w-][x-][r-][w-][x-][r-][w-][x-]$", v):
			self.userR = v[0] == "r"
			self.userW = v[1] == "w"
			self.userX = v[2] == "x"
			self.groupR = v[3] == "r"
			self.groupW = v[4] == "w"
			self.groupX = v[5] == "x"
			self.otherR = v[6] == "r"
			self.otherW = v[7] == "w"
			self.otherX = v[8] == "x"
			return self

		if re.match("^-[r-][w-][x-][r-][w-][x-][r-][w-][x-]$", v):
			self.userR = v[1] == "r"
			self.userW = v[2] == "w"
			self.userX = v[3] == "x"
			self.groupR = v[4] == "r"
			self.groupW = v[5] == "w"
			self.groupX = v[6] == "x"
			self.otherR = v[7] == "r"
			self.otherW = v[8] == "w"
			self.otherX = v[9] == "x"
			return self

		if re.match("^[ugoa]*[=][rwx]{1-3}$", v):
			self.modify(v)
			return self

		raise Exception("Unrecognized string!")
	#

	#
	# Modifies the current object data.
	# An exception is thrown if the specified string could not be parsed.
	#
	# @param	str s			A string to parse. You should specify something like "a+rw", "ug=rwx" or "o-x" here.
	#
	def modify(self, s:str):
		for sPart in s.split(","):
			sPart = sPart.strip()
			m = re.match("^([ugoa]+)([+-=])([rwx]+)$", sPart)
			if m:
				keys = []
				for t in m.group(1):
					if t == "a":
						for rwx in m.group(3):
							keys.append("user" + rwx.upper())
							keys.append("group" + rwx.upper())
							keys.append("other" + rwx.upper())
					elif t == "u":
						for rwx in m.group(3):
							keys.append("user" + rwx.upper())
					elif t == "g":
						for rwx in m.group(3):
							keys.append("group" + rwx.upper())
					else:
						for rwx in m.group(3):
							keys.append("other" + rwx.upper())

				if m.group(2) == "+":
					for key in keys:
						setattr(self, key, True)
				elif m.group(2) == "-":
					for key in keys:
						setattr(self, key, False)
				else:
					self.userR = False
					self.userW = False
					self.userX = False
					self.userR = False
					self.userW = False
					self.userX = False
					self.otherR = False
					self.otherW = False
					self.otherX = False
					for key in keys:
						setattr(self, key, True)
			else:
				raise Exception("Unrecognized string!")

		return self
	#

	def toInt(self) -> int:
		v = 0
		if self.userR:
			v = self.__setBit(v, 8)
		if self.userW:
			v = self.__setBit(v, 7)
		if self.userX:
			v = self.__setBit(v, 6)
		if self.groupR:
			v = self.__setBit(v, 5)
		if self.groupW:
			v = self.__setBit(v, 4)
		if self.groupX:
			v = self.__setBit(v, 3)
		if self.otherR:
			v = self.__setBit(v, 2)
		if self.otherW:
			v = self.__setBit(v, 1)
		if self.otherX:
			v = self.__setBit(v, 0)
		return v
	#

	#
	# Convert to octal, e.g. "755"
	#
	def toStrOcal(self) -> str:
		v = self.toInt()
		s = oct(v)
		assert s.startswith("0o")
		s = s[2:]
		assert len(s) <= 3
		while len(s) < 3:
			s = "0" + s
		return s
	#

	################################################################
	## Public Static Methods
	################################################################

	#
	# Construct a ChModValue object.
	# This method is *not* tolerant to <c>None</c>.
	#
	# @return		int|str|ChModValue v	(required) The value to parse. If <c>None</c> is specified an exception is thrown.
	# @return		ChModValue|None			Returns a ChModValue object.
	#
	@staticmethod
	def create(v):
		if isinstance(v, ChModValue):
			return v
		elif isinstance(v, (int,str)):
			return ChModValue(v)
		else:
			raise Exception("Unexpected value: " + repr(v))
	#

	#
	# Construct a ChModValue object.
	# This method is tolerant to <c>None</c>.
	#
	# @return		int|str|ChModValue v	(optional) The value to parse. If <c>None</c> is specified <c>None</c> is returned.
	# @return		ChModValue|None			Returns a ChModValue object if <c>v</c> had a non-None-value.
	#
	@staticmethod
	def createN(v):
		if v is None:
			return None
		if isinstance(v, ChModValue):
			return v
		elif isinstance(v, (int,str)):
			return ChModValue(v)
		else:
			raise Exception("Unexpected value: " + repr(v))
	#

#




#
# This code is copied from: https://gist.github.com/mkropat/7559409
# Author: mkropat
# Liense: MIT
#



import os

class UserNotFound(Exception): pass

### Windows ###
if os.name == 'nt':
	import ctypes
	from ctypes import windll, wintypes

	class WinError:	 # [1]
		InvalidParameter = 87
		InvalidSid = 1337
	#

	class NtStatus:	 # [2]
		Success		 = 0x00000000
		NoneMapped	  = 0xC0000073
	#

	class AccessMask:   # [3]
		Delete					  = 0x00010000
		ReadControl				 = 0x00020000
		WriteDAC					= 0x00040000
		WriteOwner				  = 0x00080000
		Synchronize				 = 0x00100000
		StandardRightsRequired	  = 0x000F0000
		StandardRightsRead		  = ReadControl
		StandardRightsWrite		 = ReadControl
		StandardRightsExecute	   = ReadControl
		StandardRightsAll		   = 0x001F0000
		SpecificRightsAll		   = 0x0000FFFF
	#

	class TokenAccessLevels:		# See [4] and TokenAccessLevels.cs in Mono source
		AssignPrimary	   = 1<<0
		Duplicate		   = 1<<1
		Impersonate		 = 1<<2
		Query			   = 1<<3
		QuerySource		 = 1<<4
		AdjustPrivileges	= 1<<5
		AdjustGroups		= 1<<6
		AdjustDefault	   = 1<<7
		AdjustSessionId	 = 1<<8
		Read				= AccessMask.StandardRightsRead | Query
		Write			   = AccessMask.StandardRightsWrite | AdjustPrivileges | AdjustGroups | AdjustDefault
		AllAccess		   = AccessMask.StandardRightsRequired | 0x1FF
		MaximumAllowed	  = 0x2000000
	#

	class TokenInformationClass:	# [5]
		User,\
		Groups,\
		Privileges,\
		Owner,\
		PrimaryGroup,\
		DefaultDacl,\
		Source,\
		Type,\
		ImpersonationLevel,\
		Statistics,\
		RestrictedSids,\
		SessionId,\
		GroupsAndPrivileges,\
		SessionReference,\
		SandBoxInert,\
		AuditPolicy,\
		Origin,\
		ElevationType,\
		LinkedToken,\
		Elevation,\
		HasRestrictions,\
		AccessInformation,\
		VirtualizationAllowed,\
		VirtualizationEnabled,\
		IntegrityLevel,\
		UIAccess,\
		MandatoryPolicy,\
		LogonSid,\
		IsAppContainer,\
		Capabilities,\
		AppContainerSid,\
		AppContainerNumber,\
		UserClaimAttributes,\
		DeviceClaimAttributes,\
		RestrictedUserClaimAttributes,\
		RestrictedDeviceClaimAttributes,\
		DeviceGroups,\
		RestrictedDeviceGroups,\
		SecurityAttributes,\
		IsRestricted,\
		= range(1, 41)
	#

	class SidNameUse:   # [6]
		TypeUser,\
		TypeGroup,\
		TypeDomain,\
		TypeAlias,\
		TypeWellKnownGroup,\
		TypeDeletedAccount,\
		TypeInvalid,\
		TypeUnknown,\
		TypeComputer,\
		TypeLabel,\
		= range(1, 11)
	#

	# Type declarations
	pSID = wintypes.HANDLE
	ENUM = ctypes.c_uint

	class SidAndAttributes(ctypes.Structure):   # [7]
		_fields_ = (
			('Sid', pSID),
			('Attributes', wintypes.DWORD),
		)
	#

	class TokenUser(ctypes.Structure):		  # [8]
		_fields_ = (
			('User', SidAndAttributes),
		)
	#

	_GetCurrentProcess = windll.kernel32.GetCurrentProcess	  # [9]
	_GetCurrentProcess.restype = wintypes.HANDLE

	_OpenProcessToken = windll.advapi32.OpenProcessToken		# [10]
	_OpenProcessToken.restype = wintypes.BOOL
	_OpenProcessToken.argtypes = (
		wintypes.HANDLE,				# ProcessHandle
		wintypes.DWORD,				 # DesiredAccess
		ctypes.POINTER(wintypes.HANDLE) # TokenHandle (out)
	)

	_GetTokenInformation = windll.advapi32.GetTokenInformation  # [11]
	_GetTokenInformation.restype = wintypes.BOOL
	_GetTokenInformation.argtypes = (
		wintypes.HANDLE,	# TokenHandle
		ENUM,			   # TokenInformationClass
		ctypes.c_void_p,	# TokenInformation (out)
		wintypes.DWORD,	 # TokenInformationLength
		wintypes.PDWORD	 # ReturnLength (out)
	)

	_ConvertStringSidToSid = windll.advapi32.ConvertStringSidToSidW	 # [12]
	_ConvertStringSidToSid.restype = wintypes.BOOL
	_ConvertStringSidToSid.argtypes = (
		wintypes.LPWSTR,		# StringSid
		ctypes.POINTER(pSID),   # Sid (out)
	)

	_CovertSidToStringSid = windll.advapi32.ConvertSidToStringSidW	  # [13]
	_CovertSidToStringSid.restype = wintypes.BOOL
	_CovertSidToStringSid.argtypes = (
		pSID,						   # Sid
		ctypes.POINTER(wintypes.LPWSTR) # StringSid (out)
	)

	_LocalFree = windll.kernel32.LocalFree	  # [14]
	_LocalFree.restype = None
	_LocalFree.argtypes = (wintypes.HANDLE,)

	_IsValidSid = windll.advapi32.IsValidSid	# [15]
	_IsValidSid.restype = wintypes.BOOL
	_IsValidSid.argtypes = (pSID,)

	_LookupAccountSid = windll.advapi32.LookupAccountSidW   # [16]
	_LookupAccountSid.restype = wintypes.BOOL
	_LookupAccountSid.argtypes = (
		wintypes.LPWSTR,		# lpSystemName (optional)
		pSID,				   # lpSid
		wintypes.LPWSTR,		# lpName (out, optional)
		wintypes.LPDWORD,	   # cchName (in/out)
		wintypes.LPWSTR,		# lpReferencedDomainName (out, optional)
		wintypes.LPDWORD,	   # cchReferencedDomainName (in/out)
		ctypes.POINTER(ENUM),   # peUse (out)
	)

	class Sid:
		def __init__(self, sid):
			if isinstance(sid, str):
				self.pointer = pSID()
				status = _ConvertStringSidToSid(sid, self.pointer)
				if status == WinError.InvalidSid:
					raise ValueError('Not a valid sid: %s' % sid)
				elif not status:
					raise ctypes.WinError()
			else:
				if not _IsValidSid(sid):
					raise ValueError('Invalid sid: %s' % uid)
				self.pointer = sid

		def __str__(self):
			buf = wintypes.LPWSTR()
			if not _CovertSidToStringSid(self.pointer, buf):
				raise ctypes.WinError()
			string_sid = buf.value
			_LocalFree(buf)
			return string_sid
	#

	def _getpsid():
		proc_token = wintypes.HANDLE()
		if not _OpenProcessToken(_GetCurrentProcess(), TokenAccessLevels.Read, proc_token):
			raise ctypes.WinError()

		# Get required length by calling with a 0 length buffer
		length = wintypes.DWORD()
		_GetTokenInformation(proc_token, TokenInformationClass.User, None, 0, length)

		buf = ctypes.create_string_buffer(length.value)
		if not _GetTokenInformation(proc_token, TokenInformationClass.User, buf, len(buf), length):
			raise ctypes.WinError()

		token_user = ctypes.cast(buf, ctypes.POINTER(TokenUser)).contents
		return token_user.User.Sid
	#


	def getuid() -> int:
		return str(Sid(_getpsid()))
	#

	def lookup_username(uid=None) -> str:
		sid = Sid(uid or _getpsid())

		use = ENUM()
		cchName = wintypes.DWORD()
		cchReferencedDomainName = wintypes.DWORD()

		# Get required lengths by calling with 0 length buffers
		_LookupAccountSid(None, sid.pointer, None, cchName, None, cchReferencedDomainName, use)

		name = ctypes.create_unicode_buffer(cchName.value)
		domain = ctypes.create_unicode_buffer(cchReferencedDomainName.value)

		status = _LookupAccountSid(None, sid.pointer, name, cchName, domain, cchReferencedDomainName, use)
		if status == NtStatus.NoneMapped:
			raise UserNotFound('No such sid: %s' % sid)
		elif not status:
			raise ctypes.WinError()

		if use.value != SidNameUse.TypeUser:
			raise ValueError('Not a user sid: %s' % sid)

		return name.value
	#

### Posix ###
else:
	import pwd
	from os import getuid

	def lookup_username(uid=None) -> str:
		if uid is None:
			uid = getuid()
		try:
			return pwd.getpwuid(uid)[0]
		except (KeyError, TypeError):
			raise UserNotFound('No such uid: %s' % uid)
	#



def getUserHome() -> str:
	return os.path.expanduser("~")
#










# References:
# [1] http://msdn.microsoft.com/en-us/library/windows/desktop/ms681381
# [2] http://msdn.microsoft.com/en-us/library/cc704588
# [3] http://msdn.microsoft.com/en-us/library/windows/desktop/aa374892
# [4] http://msdn.microsoft.com/en-us/library/aa374905
# [5] http://msdn.microsoft.com/en-us/library/windows/desktop/aa379626
# [6] http://msdn.microsoft.com/en-us/library/windows/desktop/aa379601
# [7] http://msdn.microsoft.com/en-us/library/windows/desktop/aa379595
# [8] http://msdn.microsoft.com/en-us/library/windows/desktop/aa379634
# [9] http://msdn.microsoft.com/en-us/library/windows/desktop/ms683179
# [10] http://msdn.microsoft.com/en-us/library/aa379295
# [11] http://msdn.microsoft.com/en-us/library/windows/desktop/aa446671
# [12] http://msdn.microsoft.com/en-us/library/aa376402
# [13] http://msdn.microsoft.com/en-us/library/windows/desktop/aa376399
# [14] http://msdn.microsoft.com/en-us/library/windows/desktop/aa366730
# [15] http://msdn.microsoft.com/en-us/library/windows/desktop/aa379151
# [16] http://msdn.microsoft.com/library/windows/desktop/aa379166






#!/usr/bin/python3




class _LockCtx(object):

	__slots__ = ( "__evt", "__bPreventFiringOnUnlock" )

	def __init__(self, evt, bPreventFiringOnUnlock:bool = False):
		self.__evt = evt
		self.__bPreventFiringOnUnlock = bPreventFiringOnUnlock
	#

	def dump(self):
		print("_LockCtx(")
		print("\t__evt = " + str(self.__evt))
		print("\t__bPreventFiringOnUnlock = " + str(self.__bPreventFiringOnUnlock))
		print(")")
	#

	def preventFiringOnUnlock(self):
		self.__bPreventFiringOnUnlock = True
	#

	def __enter__(self):
		return self
	#

	def __exit__(self, exType, exObj, traceback):
		self.__evt.unlock(bPreventFiring = self.__bPreventFiringOnUnlock or (exType is not None))
	#

#




class ObservableEvent(object):

	def __init__(self, name:str = None, bCatchExceptions:bool = False):
		self.__name = name
		self.__listeners = tuple()
		self.__bCatchExceptions = bCatchExceptions
		self.__lockCounter = 0
		self.__bNeedFiring = False
		self.__lockCtx = None
	#

	@property
	def catchExceptions(self):
		return self.__bCatchExceptions
	#

	@property
	def name(self):
		return self.__name
	#

	@property
	def listeners(self):
		return self.__listeners
	#

	def __len__(self):
		return len(self.__listeners)
	#

	def removeAllListeners(self):
		self.__listeners = tuple()
	#

	def __str__(self):
		if self.__name:
			ret = repr(self.__name)[1:][:-1] + "("
		else:
			ret = "Event("

		if len(self.__listeners) == 0:
			ret += "no listener"
		elif len(self.__listeners) == 1:
			ret += "1 listener"
		else:
			ret += str(len(self.__listeners)) + " listeners"

		if self.__lockCounter > 0:
			ret += ", lockCounter=" + str(self.__lockCounter)
			if self.__bNeedFiring:
				ret += ", fire on unlock"

		return ret + ")"
	#

	def __repr__(self):
		return self.__str__()
	#

	def add(self, theCallable:callable):
		assert theCallable != None
		self.__listeners += (theCallable,)
		return self
	#

	def __iadd__(self, theCallable:callable):
		assert theCallable != None
		self.__listeners += (theCallable,)
		return self
	#

	def remove(self, theCallable:callable) -> bool:
		assert theCallable != None
		if theCallable in self.__listeners:
			n = self.__listeners.index(theCallable)
			self.__listeners = self.__listeners[:n] + self.__listeners[n + 1:]
			return True
		else:
			return False
	#

	def __isub__(self, theCallable:callable):
		assert theCallable != None
		if theCallable in self.__listeners:
			n = self.__listeners.index(theCallable)
			self.__listeners = self.__listeners[:n] + self.__listeners[n + 1:]
			return True
		else:
			return False
	#

	def __call__(self, *argv, **kwargs):
		if self.__bCatchExceptions:
			try:
				for listener in self.__listeners:
					listener(*argv, **kwargs)
			except Exception as ee:
				pass
		else:
			for listener in self.__listeners:
				listener(*argv, **kwargs)
	#

	def fire(self, *argv, **kwargs):
		if self.__lockCounter > 0:
			self.__bNeedFiring = True

		else:
			self.__bNeedFiring = False

			if self.__bCatchExceptions:
				try:
					for listener in self.__listeners:
						listener(*argv, **kwargs)
				except Exception as ee:
					pass
			else:
				for listener in self.__listeners:
					listener(*argv, **kwargs)
	#

	def lock(self):
		self.__lockCounter += 1
		if self.__lockCtx is None:
			self.__lockCtx = _LockCtx(self)
		return self.__lockCtx
	#

	def unlock(self, bPreventFiring:bool = False):
		assert self.__lockCounter > 0
		self.__lockCounter -= 1
		if self.__bNeedFiring:
			if bPreventFiring:
				self.__bNeedFiring = False
			else:
				self.fire()
	#

#



import os
import json




#
# Maybe use https://stackoverflow.com/questions/18406165/creating-a-timer-in-python/18406263
# in order to delay writes for 2 seconds?
#




class PersistentProperties(object):

	def __init__(self, filePath:str, bAutoStore:bool = True):
		self.__filePath = filePath
		self.__data = self.__tryLoad()
		self.__bChanged = False
		self.__bAutoStore = bAutoStore
	#

	@property
	def isAutoStore(self):
		return self.__bAutoStore
	#

	@property
	def isChanged(self):
		return self.__bChanged
	#

	def __tryLoad(self):
		if os.path.isfile(self.__filePath):
			with open(self.__filePath, "r") as f:
				return json.loads(f.read())
		else:
			dirPath = os.path.dirname(self.__filePath)
			if not os.path.isdir(dirPath):
				raise Exception("Won't be able to store persistent data: Directory does not exist!")
			return {}
	#

	def put(self, key:str, value):
		assert isinstance(key, str)
		assert isinstance(value, (int, float, str, type(None), list, tuple, dict))

		if key in self.__data:
			if value != self.__data[key]:
				self.__data[key] = value

				if self.__bAutoStore:
					self.__store()
				else:
					self.__bChanged = True
		else:
			self.__data[key] = value

			if self.__bAutoStore:
				self.__store()
			else:
				self.__bChanged = True
	#

	def get(self, key:str, defaultValue = None):
		assert isinstance(key, str)

		return self.__data.get(key, defaultValue)
	#

	def remove(self, key:str):
		assert isinstance(key, str)

		if key in self.__data:
			del self.__data[key]

			if self.__bAutoStore:
				self.__store()
			else:
				self.__bChanged = True
	#

	def store(self):
		if self.__bChanged:
			self.__store()
			self.__bChanged = False
	#

	def __store(self):
		with open(self.__filePath, "w") as f:
			f.write(json.dumps(self.__data))
	#

#












import os
import collections


def setFileExt(filePath:str, newFileExt:str):
	if not newFileExt.startswith("."):
		newFileExt = "." + newFileExt

	pos = filePath.rfind("/")
	if pos > 0:
		dirPath = filePath[:pos+1]
		filePath = filePath[pos+1:]
	else:
		dirPath = ""

	pos = filePath.rfind(".")
	if pos > 0:
		filePath = filePath[:pos] + newFileExt
	else:
		filePath += newFileExt

	return dirPath + filePath
#

def getFileExt(filePath:str):
	pos = filePath.rfind("/")
	if pos > 0:
		dirPath = filePath[:pos+1]
		filePath = filePath[pos+1:]
	else:
		dirPath = ""

	pos = filePath.rfind(".")
	if pos > 0:
		return filePath[:pos + 1]
	else:
		return None
#

def makeAbsDirPathAndCheckDirExists(baseDir:str, dirPath:str):
	if baseDir:
		if not os.path.isdir(baseDir):
			raise Exception("Base directory does not exist: " + repr(baseDir))
	if not os.path.isabs(dirPath):
		if baseDir is None:
			dirPath = os.path.abspath(dirPath)
		else:
			dirPath = os.path.abspath(os.path.join(baseDir, dirPath))
	if not os.path.isdir(dirPath):
		raise Exception("Directory does not exist: " + repr(dirPath))
	return dirPath
#

def makeAbsFilePathAndCheckFileExists(baseDir:str, filePath:str):
	if baseDir:
		if not os.path.isdir(baseDir):
			raise Exception("Base directory does not exist: " + repr(baseDir))
	if not os.path.isabs(filePath):
		if baseDir is None:
			filePath = os.path.abspath(filePath)
		else:
			filePath = os.path.abspath(os.path.join(baseDir, filePath))
	if not os.path.isfile(filePath):
		raise Exception("File does not exist: " + repr(filePath))
	return filePath
#

def makeAbsFilePathAndCheckBaseDirExists(baseDir:str, filePath:str):
	if baseDir:
		if not os.path.isdir(baseDir):
			raise Exception("Base directory does not exist: " + repr(baseDir))
	if not os.path.isabs(filePath):
		if baseDir is None:
			filePath = os.path.abspath(filePath)
		else:
			filePath = os.path.abspath(os.path.join(baseDir, filePath))
	baseDir = os.path.dirname(filePath)
	if not os.path.isfile(baseDir):
		raise Exception("Base directory of file does not exist: " + repr(filePath))
	return filePath
#











import typing
import re

from .showCapacityProgress import formatBytes





#
# This class represents a specific amount of bytes. An object of this class is intended to either represent an existing amount of data specified by users,
# by a configuration file or represent data that requires human readable formatting.
#
class AmountOfBytes(object):

	################################################################################################################################
	## Constructor
	################################################################################################################################

	def __init__(self, value):
		self.__n = AmountOfBytes.__parseBytesFromAny(value)
	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	def __repr__(self):
		return "AmountOfBytes<(" + AmountOfBytes.__bytesToStrAuto(self.__n) + ")>"
	#

	def __str__(self):
		return AmountOfBytes.__bytesToStrAuto(self.__n)
	#

	#
	# Use this method to retrieve formatted output of the data.
	#
	def toStr(self, bShort:bool = True, magnitude:str = None) -> str:
		if magnitude in "KBMGT":
			return AmountOfBytes.__bytesToStrFixed(self.__n, magnitude)
		elif bShort:
			return formatBytes(self.__n)
		else:
			return AmountOfBytes.__bytesToStrAuto(self.__n)
	#

	def __float__(self):
		return float(self.__n)
	#

	def __int__(self):
		return self.__n
	#

	def __add__(self, other):
		if isinstance(other, AmountOfBytes):
			return AmountOfBytes(self.__n + other.__n)
		elif isinstance(other, int):
			return AmountOfBytes(self.__n + other)
		elif isinstance(other, float):
			return self.__n + other
		else:
			raise TypeError(repr(other))
	#

	def __sub__(self, other):
		if isinstance(other, AmountOfBytes):
			return AmountOfBytes(self.__n - other.__n)
		elif isinstance(other, int):
			return AmountOfBytes(self.__n - other)
		elif isinstance(other, float):
			return self.__n - other
		else:
			raise TypeError(repr(other))
	#

	def __mul__(self, other):
		if isinstance(other, AmountOfBytes):
			raise TypeError(repr(other))
		elif isinstance(other, int):
			return AmountOfBytes(self.__n * other)
		elif isinstance(other, float):
			return self.__n * other
		else:
			raise TypeError(repr(other))
	#

	def __div__(self, other):
		if isinstance(other, AmountOfBytes):
			raise TypeError(repr(other))
		elif isinstance(other, int):
			return self.__n / other
		elif isinstance(other, float):
			return self.__n / other
		else:
			raise TypeError(repr(other))
	#

	def __eq__(self, other):
		if isinstance(other, AmountOfBytes):
			return self.__n == other.__n
		elif isinstance(other, int):
			return self.__n == other
		else:
			return False
	#

	def __ne__(self, other):
		if isinstance(other, AmountOfBytes):
			return self.__n != other.__n
		elif isinstance(other, int):
			return self.__n != other
		else:
			return False
	#

	def __gt__(self, other):
		if isinstance(other, AmountOfBytes):
			return self.__n > other.__n
		elif isinstance(other, (int, float)):
			return self.__n > other
		else:
			return False
	#

	def __ge__(self, other):
		if isinstance(other, AmountOfBytes):
			return self.__n >= other.__n
		elif isinstance(other, (int, float)):
			return self.__n >= other
		else:
			return False
	#

	def __lt__(self, other):
		if isinstance(other, AmountOfBytes):
			return self.__n < other.__n
		elif isinstance(other, (int, float)):
			return self.__n < other
		else:
			return False
	#

	def __le__(self, other):
		if isinstance(other, AmountOfBytes):
			return self.__n <= other.__n
		elif isinstance(other, (int, float)):
			return self.__n <= other
		else:
			return False
	#

	################################################################################################################################
	## Static Helper Methods
	################################################################################################################################

	@staticmethod
	def __parseBytesFromAny(v:typing.Union[int,str]) -> int:
		if isinstance(v, int):
			assert v >= 0
			return v

		vText = v.strip().upper()

		m = re.match("^([0-9.]+)\s*([KMGT]B?)$", vText)
		if m:
			try:
				n = float(m.group(1))
			except:
				raise Exception("Value does not specify bytes: " + repr(vText))

			sFactor = m.group(2)
			if sFactor.startswith("K"):
				n *= 1024
			elif sFactor.startswith("M"):
				n *= 1024*1024
			elif sFactor.startswith("G"):
				n *= 1024*1024*1024
			elif sFactor.startswith("T"):
				n *= 1024*1024*1024*1024
			else:
				pass

			return int(n)

		else:
			try:
				return int(vText)
			except Exception as ee:
				raise Exception("Value does not specify bytes: " + repr(vText))
	#

	@staticmethod
	def __bytesToStrAuto(n:int) -> str:
		if (n >= 1024*1024*1024*1024) and ((n % (1024*1024*1024*1024)) == 0):
			return str(n // (1024*1024*1024*1024)) + "T"
		elif (n >= 1024*1024*1024) and ((n % (1024*1024*1024)) == 0):
			return str(n // (1024*1024*1024)) + "G"
		elif (n >= 1024*1024) and ((n % (1024*1024)) == 0):
			return str(n // (1024*1024)) + "M"
		elif (n >= 1024) and ((n % 1024) == 0):
			return str(n // 1024) + "K"
		else:
			return str(n)
	#

	@staticmethod
	def __bytesToStrFixed(n:int, mag:str) -> str:
		if mag == "T":
			v = round(n / (1024*1024*1024*1024), 2)
		elif mag == "G":
			v = round(n / (1024*1024*1024), 2)
		elif mag == "M":
			v = round(n / (1024*1024), 2)
		elif mag == "K":
			v = round(n / 1024, 1)
		else:
			v = n

		return str(v) + " " + mag
	#

	################################################################################################################################
	## Public Static Methods
	################################################################################################################################

	@staticmethod
	def parseFromStr(v:str):
		assert isinstance(v, str)
		return AmountOfBytes(v)
	#

	@staticmethod
	def parse(v:typing.Union[int,str]):
		assert isinstance(v, (int,str))
		return AmountOfBytes(v)
	#

#



















import enum





#
# This is a base class for enumerations.
#
class EnumBase(enum.Enum):

	def __new__(cls, value, name):
		member = object.__new__(cls)
		member._value_ = value
		member.fullname = name
		return member
	#

	#
	# Get an integer representation of this enumeration item.
	#
	def __int__(self):
		return self._value_
	#

	#
	# Get a string representation of this enumeration item.
	#
	def __str__(self):
		return self.fullname
	#

	#
	# Get a string representation of this enumeration item.
	#
	def __repr__(self):
		return self.fullname
	#

	#
	# Get an integer representation of this enumeration item.
	#
	def toJSON(self):
		return self._value_
	#

	#
	# Get a list of all states this enumeration contains.
	#
	@classmethod
	def allStates(cls):
		ret = []
		for key in cls.__dict__["_value2member_map_"]:
			enumItem = cls.__dict__["_value2member_map_"][key]
			ret.append(enumItem)
		return ret
	#

	#
	# This method converts a string or integer representing an enumeration value to an actual enumeration value.
	#
	# @param		mixed data							Either a string or an integer to parse. (A member of this enumeration is
	#													accepted as well and passed through to the caller as in this case there
	#													is no need to parse anything.)
	# @param		bool bRaiseExceptionOnError			If <c>True</c> (which is the default) an exception is thrown if a
	#													value has been specified that could not be parsed. (Please note that
	#													an exception is always thrown if the spcified value is not a string nor
	#													an integer nor of the enumeration type itself.)
	#
	@classmethod
	def parse(cls, data, bRaiseExceptionOnError:bool = True):
		if isinstance(data, int):
			if data in cls.__dict__["_value2member_map_"]:
				return cls.__dict__["_value2member_map_"][data]
			else:
				if bRaiseExceptionOnError:
					raise Exception("Not a member of enumeration '" + str(cls.__name__) + "': " + repr(data))
				else:
					return None
		elif isinstance(data, str):
			for key in cls.__dict__["_value2member_map_"]:
				enumItem = cls.__dict__["_value2member_map_"][key]
				if str(enumItem) == data:
					return enumItem
			if data in cls.__dict__["_member_names_"]:
				return cls.__dict__[data]
			else:
				if bRaiseExceptionOnError:
					raise Exception("Not a member of enumeration '" + str(cls.__name__) + "': " + repr(data))
				else:
					return None
		elif cls == type(data):
			return data
		else:
			raise Exception("Unrecognized enumeration value type: " + repr(data))
	#

#












class CmdLineParser(object):

	@staticmethod
	def parseCmdLine(text):

		IN_SPACE = 1
		IN_SPACE_EXPECTS_SPACE = 2
		IN_WORD = 3
		IN_WORD_NEXT_MASKED = 4
		IN_STR = 5
		IN_STR_NEXT_MASKED = 6

		ret = []
		mode = IN_SPACE
		buffer = []
		charNo = 0

		for c in text:
			charNo += 1

			if mode == IN_SPACE:
				if (c == " ") or (c == "\t"):
					continue
				elif c == "\"":
					mode = IN_STR
				else:
					buffer.append(c)
					mode = IN_WORD
			elif mode == IN_SPACE_EXPECTS_SPACE:
				if (c == " ") or (c == "\t"):
					mode = IN_SPACE
				else:
					raise Exception("Space character expected at position " + str(charNo) + ": " + c + " (" + str(ord(c)) + ")")
			elif mode == IN_WORD:
				if (c == " ") or (c == "\t"):
					ret.append("".join(buffer))
					buffer.clear()
					mode = IN_SPACE
				elif c == "\\":
					mode = IN_WORD_NEXT_MASKED
				elif c == "\"":
					raise Exception("Unexpected character at position " + str(charNo) + ": " + c + " (" + str(ord(c)) + ")")
				else:
					buffer.append(c)
			elif mode == IN_WORD_NEXT_MASKED:
				buffer.append(c)
				mode = IN_WORD
			elif mode == IN_STR:
				if c == "\"":
					ret.append("".join(buffer))
					buffer.clear()
					mode = IN_SPACE_EXPECTS_SPACE
				elif c == "\\":
					mode = IN_STR_NEXT_MASKED
				else:
					buffer.append(c)
			elif mode == IN_STR_NEXT_MASKED:
				buffer.append(c)
				mode = IN_STR
			else:
				raise Exception()

		if mode == IN_WORD:
			ret.append("".join(buffer))
		elif mode == IN_STR:
			raise Exception("Unterminated string at position " + str(charNo) + ": " + c + " (" + str(ord(c)) + ")")
		elif mode == IN_STR_NEXT_MASKED:
			raise Exception("Unterminated string at position " + str(charNo) + ": " + c + " (" + str(ord(c)) + ")")

		return ret
	#


#











import os
import sys
import datetime





class dtutils(object):

	__EPOCH = datetime.datetime.utcfromtimestamp(0)



	@staticmethod
	def epochTimeStamp():
		return dtutils.__EPOCH

	#



	@staticmethod
	def dateTimeToSecondsSinceEpoch(dt):
		return (dt - dtutils.__EPOCH).total_seconds()

	#



	@staticmethod
	def dateTimeToMillisecondsSinceEpoch(dt):
		return (dt - dtutils.__EPOCH).total_seconds() * 1000

	#



	@staticmethod
	def secondsSinceEpochToDateTime(seconds):
		datetime.datetime.utcfromtimestamp(seconds)

	#



	@staticmethod
	def nowInSecondsSinceEpoch():
		return (datetime.datetime.utcnow() - dtutils.__EPOCH).total_seconds()

	#



	@staticmethod
	def nowInMillisecondsSinceEpoch():
		return int((datetime.datetime.utcnow() - dtutils.__EPOCH).total_seconds() * 1000)

	#










class MultiCounterDict(dict):

	def __init__(self):
		pass
	#

	def increment(self, *keys):
		length = len(keys)
		dParent = self
		for i in range(0, length):
			bIsLast = i == (length - 1)
			key = keys[i]

			if bIsLast:
				dParent[key] = dParent.get(key, 0) + 1
			else:
				d = dParent.get(key, None)
				if d is None:
					d = {}
					dParent[key] = d
				dParent = d
	#

#







#
# This class wraps around a value. It contains not only the value itself but type information as well.
#
class TypedValue(object):

	def __init__(self, dataType, value):
		self.__dataType = dataType
		self.__value = value
	#

	@property
	def value(self):
		return self.__value
	#

	@property
	def dataType(self):
		return self.__dataType
	#

	def __str__(self):
		if isinstance(self.__value, str):
			return "V(" + self.__dataType + ": \"" + self.__value.replace("\"", "\\\"") + "\")"
		else:
			return "V(" + self.__dataType + ": " + str(self.__value) + ")"
	#

	def __repr__(self):
		return self.__str__()
	#

	def __eq__(self, other):
		if isinstance(other, TypedValue):
			return (other.__dataType == self.__dataType) and (other.__value == self.__value)
		else:
			return False
	#

	def __ne__(self, other):
		if isinstance(other, TypedValue):
			return (other.__dataType != self.__dataType) or (other.__value != self.__value)
		else:
			return True
	#

#














import re
import collections
import typing

from .Token import Token
from .TokenizationError import TokenizationError






class _TokenizationState(object):

	def __init__(self, initialState:str):
		self.lineNo = 1
		self.line_start = 0
		self.pos = 0
		self.currentState = initialState
	#

#




#
# RegEx based tokenizer: This tokenizer uses regular expressions in order to create tokens.
#
class RegExBasedTokenizingTable(object):

	#
	# Initialization method
	#
	# @param		object[] patternDefs	A list of pattern definitions. Each pattern definition should be a tuple or list of one of the following structures:
	#										* 4 items
	#											* (required) str: the token type to use on match
	#											* (required) str: the pattern to match
	#											* (optional) str: next state
	#											* (optional) callable: parsing delegate
	#										* 6 items
	#											* (required) str: the token type to use on match
	#											* (optional) str: if not None a pattern to match but that will not be part of the token content
	#											* (required) str: the pattern to match (which will be the content of the token)
	#											* (optional) str: if not None a pattern to match but that will not be part of the token content
	#											* (optional) str: next state
	#											* (optional) callable: parsing delegate
	#
	def __init__(self, stateName:str, patternDefs, bDebug:bool = False):
		assert isinstance(stateName, str)
		assert stateName

		assert isinstance(patternDefs, (tuple, list))
		assert len(patternDefs) > 0

		self.__stateName = stateName

		self.__tokenPseudoTypeToTokenTypeMap = {
			"NEWLINE": "NEWLINE",
			"SPACE": "SPACE",
			"ERROR": "ERROR",
		}

		self.__pattern2StateMap = {}
		self.__tokenParsingDelegatesMap = {}
		rawPatternsList = []
		rawPatternsSources = []
		n = 1
		for patternDef in patternDefs:
			assert isinstance(patternDef, (tuple, list))

			if len(patternDef) == 4:
				tokenType, pattern, nextState, parsingDelegate = patternDef
				patternPre = None
				patternPost = None
			elif len(patternDef) == 6:
				tokenType, patternPre, pattern, patternPost, nextState, parsingDelegate = patternDef
			else:
				raise Exception("Invalid pattern definition: " + str(patternDef))
			tokenPseudoType = tokenType + "_" + str(n)
			self.__tokenPseudoTypeToTokenTypeMap[tokenPseudoType] = tokenType

			p2 = []
			if patternPre:
				p2.extend(("(?:", patternPre, ")"))
			p2.extend(("(?P<", tokenPseudoType, ">", pattern, ")"))
			if patternPost:
				p2.extend(("(?:", patternPost, ")"))

			rawPatternsList.append("".join(p2)),
			rawPatternsSources.append(tokenType)
			self.__pattern2StateMap[tokenPseudoType] = nextState
			self.__tokenParsingDelegatesMap[tokenPseudoType] = parsingDelegate

			n += 1

		self.__compiledPatterns = None

		if bDebug:
			print("Compiling regular expressions for " + self.__class__.__name__ + " state " + repr(stateName) + ":")
			for tt, r in zip(rawPatternsSources, rawPatternsList):
				print("\t" + repr(tt) + "\t\t=>  " + repr(r))

		rawPatterns = "|".join(rawPatternsList)
		rawPatterns += "|(?P<NEWLINE>\n)"
		rawPatterns += "|(?P<SPACE>[\t ]+)"
		rawPatterns += "|(?P<ERROR>.)"
		try:
			self.__compiledPatterns = re.compile(rawPatterns)
			if bDebug:
				print("\tPATTERN: " + rawPatterns)
		except Exception as e:
			print("==== COMPILE ERROR")
			print("==== " + e.msg)
			print("==== " + rawPatterns)
			s = ""
			for i in range(0, e.colno):
				s += " "
			s += "^"
			print(s)
			raise
	#



	def addStateTransition(self, tokenType:str, nextState:str):
		self.__pattern2StateMap[tokenType] = nextState
		return self
	#



	@property
	def stateName(self):
		return self.__stateName
	#



	def _getNextToken(self, state:_TokenizationState, text:str) -> tuple:
		mo = self.__compiledPatterns.match(text, state.pos)
		columnNo = state.pos - state.line_start + 1
		if not mo:
			raise TokenizationError("No RegEx match", state.lineNo, columnNo, state.currentState)
		if state.pos != mo.start():
			raise TokenizationError("Invalid RegEx match pos", state.lineNo, columnNo, state.currentState)

		tokenPseudoType = mo.lastgroup
		value = mo.group(tokenPseudoType)
		tokenType = self.__tokenPseudoTypeToTokenTypeMap[tokenPseudoType]
		if tokenType == "NEWLINE":
			ret = Token(tokenType, value, state.lineNo, columnNo, mo.end() - state.pos)
			state.line_start = mo.end()
			state.lineNo += 1
		elif tokenType == "SPACE":
			ret = Token(tokenType, value, state.lineNo, columnNo, mo.end() - state.pos)
		elif tokenType == "ERROR":
			raise TokenizationError("Error match", state.lineNo, columnNo, state.currentState)
		else:
			d = self.__tokenParsingDelegatesMap.get(tokenPseudoType)
			if d != None:
				value = d(value)
			ret = Token(tokenType, value, state.lineNo, columnNo, mo.end() - state.pos)
			if tokenType == 'newline':
				state.line_start = mo.end()
				state.lineNo += 1

		#print(">>EMITTING:", ret)
		nextState = self.__pattern2StateMap.get(tokenPseudoType)
		if nextState:
			#print(">>SWITCHING FROM STATE", repr(state.currentState), "TO STATE", repr(nextState))
			state.currentState = nextState

		state.pos = mo.end()

		return ret
	#



	def _getAllResultStates(self) -> list:
		return [ x for x in self.__pattern2StateMap.values() if x is not None ]
	#



#




class RegExBasedTableTokenizer(object):

	def __init__(self, initialState:str, tokenizationTables:typing.Union[list,tuple,typing.Sequence]):
		assert isinstance(initialState, str)
		assert initialState

		tokenizationTables = list(tokenizationTables)
		self.__tokenizationTableMap = {}

		allStates = set()
		allResultStates = set()
		for table in tokenizationTables:
			assert isinstance(table, RegExBasedTokenizingTable)
			if table.stateName in self.__tokenizationTableMap:
				raise Exception("Duplicate table: " + table.stateName)
			for s in table._getAllResultStates():
				allResultStates.add(s)
			allStates.add(table.stateName)
			self.__tokenizationTableMap[table.stateName] = table
		#print("allStates:", allStates)
		#print("allResultStates:", allResultStates)
		wrongResultStates = allResultStates - allStates
		if wrongResultStates:
			raise Exception("No such states: " + str(sorted(wrongResultStates)))

		if initialState not in self.__tokenizationTableMap:
			raise Exception("No such initial state: " + repr(initialState))

		self.__initialState = initialState
	#



	#
	# @return		Token,str[] tokens			Returns token objects according to the pattern defined during initialization.
	#
	def tokenize(self, text, bEmitWhiteSpaces:bool = False, bEmitNewLines:bool = False):
		assert isinstance(text, str)
		assert isinstance(bEmitWhiteSpaces, bool)
		assert isinstance(bEmitNewLines, bool)

		state = _TokenizationState(self.__initialState)
		maxPos = len(text)
		while state.pos < maxPos:
			currentTable = self.__tokenizationTableMap[state.currentState]
			t = currentTable._getNextToken(state, text)
			if t.type.startswith("__"):
				continue
			elif t.type == "NEWLINE":
				if bEmitNewLines:
					yield t
			elif t.type == "SPACE":
				if bEmitWhiteSpaces:
					yield t
			else:
				yield t
	#



#









import typing
import datetime

import jk_typing






_EPOCH = datetime.datetime.utcfromtimestamp(0)

def _dateTimeToSecondsSinceEpoch(dt:datetime.datetime) -> float:
	assert isinstance(dt, datetime.datetime)
	return (dt - _EPOCH).total_seconds()
#

def _secondsSinceEpocheToDateTime(t:float) -> datetime.datetime:
	assert isinstance(t, float)
	assert t >= 0
	return datetime.datetime.utcfromtimestamp(t)
#







class TimeStamp(object):

	################################################################################################################################
	## Constructor
	################################################################################################################################

	#
	# Constructor method.
	#
	@jk_typing.checkFunctionSignature()
	def __init__(self, value):
		if isinstance(value, (int,float)):
			self.__t = value
		elif isinstance(value, datetime.datetime):
			self.__t = _dateTimeToSecondsSinceEpoch(value)
		else:
			raise Exception()
	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	@property
	def secondsSinceEpoch(self) -> float:
		return self.__t
	#

	@property
	def dateTime(self) -> datetime.datetime:
		return _secondsSinceEpocheToDateTime(self.__t)
	#

	################################################################################################################################
	## Helper Methods
	################################################################################################################################

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	def __repr__(self):
		return "TimeStamp<({})".format(self.dateTime)
	#

	def __str__(self):
		return "TimeStamp<({})".format(self.dateTime)
	#

	#
	# Use this method to retrieve formatted output of the data.
	#
	def toStr(self) -> str:
		return "{}".format(self.dateTime)
	#

	def toDateTime(self) -> datetime.datetime:
		return _secondsSinceEpocheToDateTime(self.__t)
	#

	def __float__(self):
		return self.__t
	#

	def __int__(self):
		return int(self.__t)
	#

	def __sub__(self, other):
		if isinstance(other, TimeStamp):
			return self.__t - other.__t
		elif isinstance(other, (int,float)):
			return TimeStamp(self.__t - other)
		else:
			raise TypeError(repr(other))
	#

	def __add__(self, other):
		if isinstance(other, (int,float)):
			return TimeStamp(self.__t + other)
		else:
			raise TypeError(repr(other))
	#

	def __gt__(self, other):
		tOther = None
		if isinstance(other, (int,float)):
			tOther = other
		elif isinstance(other, datetime.datetime):
			tOther = _dateTimeToSecondsSinceEpoch(other)
		elif isinstance(other, TimeStamp):
			tOther = other.__t
		else:
			raise TypeError(repr(other))

		return self.__t > tOther
	#

	def __ge__(self, other):
		tOther = None
		if isinstance(other, (int,float)):
			tOther = other
		elif isinstance(other, datetime.datetime):
			tOther = _dateTimeToSecondsSinceEpoch(other)
		elif isinstance(other, TimeStamp):
			tOther = other.__t
		else:
			raise TypeError(repr(other))

		return self.__t >= tOther
	#

	def __lt__(self, other):
		tOther = None
		if isinstance(other, (int,float)):
			tOther = other
		elif isinstance(other, datetime.datetime):
			tOther = _dateTimeToSecondsSinceEpoch(other)
		elif isinstance(other, TimeStamp):
			tOther = other.__t
		else:
			raise TypeError(repr(other))

		return self.__t < tOther
	#

	def __le__(self, other):
		tOther = None
		if isinstance(other, (int,float)):
			tOther = other
		elif isinstance(other, datetime.datetime):
			tOther = _dateTimeToSecondsSinceEpoch(other)
		elif isinstance(other, TimeStamp):
			tOther = other.__t
		else:
			raise TypeError(repr(other))

		return self.__t <= tOther
	#

	def __eq__(self, other):
		tOther = None
		if isinstance(other, (int,float)):
			tOther = other
		elif isinstance(other, datetime.datetime):
			tOther = _dateTimeToSecondsSinceEpoch(other)
		elif isinstance(other, TimeStamp):
			tOther = other.__t
		else:
			raise TypeError(repr(other))

		fDiff = abs(self.__t - tOther)
		return fDiff < 0.0000000001
	#

	def __ne__(self, other):
		tOther = None
		if isinstance(other, (int,float)):
			tOther = other
		elif isinstance(other, datetime.datetime):
			tOther = _dateTimeToSecondsSinceEpoch(other)
		elif isinstance(other, TimeStamp):
			tOther = other.__t
		else:
			raise TypeError(repr(other))

		fDiff = abs(self.__t - tOther)
		return fDiff >= 0.0000000001
	#

	################################################################################################################################
	## Public Static Methods
	################################################################################################################################

	@staticmethod
	def parse(v:typing.Union[int,float]):
		assert isinstance(v, (int,float))
		return TimeStamp(v)
	#

	@staticmethod
	def now():
		return TimeStamp(datetime.datetime.utcnow())
	#

#








from .RGB import RGB
from .RGBSpectrum import RGBSpectrum



class RGBSpectrumBuilder(object):

	def __init__(self):
		self.__parts = []
	#

	def __buildRGBSpectrumPart(self, c1:RGB, c2:RGB, length:int):
		values = []
		for i in range(0, length):
			v = i / (length -1)
			vv = 1 - v
			r = c1.r * vv + c2.r * v
			g = c1.g * vv + c2.g * v
			b = c1.b * vv + c2.b * v
			values.append(RGB(r, g, b))
		return values
	#

	def appendGradientBand(self, fromRGB:RGB, toRGB:RGB, length:int):
		self.__parts.extend(self.__buildRGBSpectrumPart(fromRGB, toRGB, length))
		return self
	#

	def insertGradientBand(self, position:int, fromRGB:RGB, toRGB:RGB, length:int):
		parts1 = self.__parts[0:position]
		parts2 = self.__parts[position:]
		self.__parts = parts1 + self.__buildRGBSpectrumPart(fromRGB, toRGB, length) + parts2
		return self
	#

	def appendFlatBand(self, rgb:RGB, length:int):
		self.__parts.extend([ rgb ] * length)
		return self
	#

	def insertFlatBand(self, position:int, rgb:RGB, length:int):
		parts1 = self.__parts[0:position]
		parts2 = self.__parts[position:]
		self.__parts = parts1 + ([ rgb ] * length) + parts2
		return self
	#

	def length(self) -> int:
		return len(self.__parts)
	#

	def __len__(self) -> int:
		return len(self.__parts)
	#

	def compile(self) -> RGBSpectrum:
		return RGBSpectrum(self.__parts)
	#

#









from .T5 import T5







def timeRange5(fromHour:T5, toHour:T5, includeRightBorder:bool = False, steps:int = 1):
	i = fromHour.clone()
	n = 0
	while i < toHour:
		if n % steps == 0:
			yield i.clone()
		i._absoluteTick += 1
		i._hour = i._absoluteTick // 12
		i._minuteTick = i._absoluteTick % 12
		i._minute = i._minuteTick * 5
		n += 1
	if i == toHour:
		if includeRightBorder:
			yield i.clone()
#











import re
import collections

from .Token import Token






#
# RegEx based tokenizer: This tokenizer uses regular expressions in order to create tokens.
#
class RegExBasedTokenizer(object):

	#
	# Initialization method
	#
	# @param		str[] patternDefs		A list of pattern definitions. Each pattern definition should be a tuple or list of one of the following structures:
	#										* 2 items
	#											* (required) the token type to use on match
	#											* (required) the pattern to match
	#										* 4 items
	#											* (required) the token type to use on match
	#											* (optional) if not None a pattern to match but that will not be part of the token content
	#											* (required) the pattern to match (which will be the content of the token)
	#											* (optional) if not None a pattern to match but that will not be part of the token content
	#
	def __init__(self, patternDefs):
		assert isinstance(patternDefs, (tuple, list))
		assert len(patternDefs) > 0

		p = ""
		for patternDef in patternDefs:
			assert isinstance(patternDef, (tuple, list))
			if len(p) > 0:
				p += "|"
			if len(patternDef) == 2:
				p += '(?P<' + patternDef[0] + '>' + patternDef[1] + ')'
			elif len(patternDef) == 4:
				if patternDef[1] != None:
					p += '(?:' + patternDef[1] + ')'
				p += '(?P<' + patternDef[0] + '>' + patternDef[2] + ')'
				if patternDef[3] != None:
					p += '(?:' + patternDef[3] + ')'
			else:
				raise Exception("Invalid pattern definition: " + str(patternDef))

		self.__rawPatterns = p
		self.__compiledPatterns = None
		self.__typeParsingDelegates = {}
	#



	@property
	def isCompiled(self):
		return self.__compiledPatterns != None
	#



	def compile(self):
		rawPatterns = self.__rawPatterns
		rawPatterns += "|(?P<NEWLINE>\n)"
		rawPatterns += "|(?P<SPACE>[\t ]+)"
		rawPatterns += "|(?P<ERROR>.)"
		try:
			self.__compiledPatterns = re.compile(rawPatterns)
		except Exception as e:
			print("==== COMPILE ERROR")
			print("==== " + e.msg)
			print("==== " + rawPatterns)
			s = ""
			for i in range(0, e.colno):
				s += " "
			s += "^"
			print(s)
			raise
	#



	def addTokenPattern(self, tokenType:str, patternRegExStr):
		assert isinstance(tokenType, str)

		bErr = False
		if isinstance(patternRegExStr, (tuple, list)):
			if len(patternRegExStr) == 1:
				if isinstance(patternRegExStr[0], str):
					patternRegExStr = patternRegExStr[0]
				else:
					bErr = True
			elif len(patternRegExStr) == 3:
				pass
			else:
				bErr = True
		elif isinstance(patternRegExStr, str):
			pass
		else:
			bErr = True
		if bErr:
			raise Exception("Invalid patternRegExStr specified!")

		if len(self.__rawPatterns) > 0:
			self.__rawPatterns += "|"
		if isinstance(patternRegExStr, str):
			self.__rawPatterns += '(?P<' + tokenType + '>' + patternRegExStr + ')'
		else:
			if patternRegExStr[0] != None:
				self.__rawPatterns += '(?:' + patternRegExStr[0] + ')'
			self.__rawPatterns += '(?P<' + tokenType + '>' + patternRegExStr[1] + ')'
			if patternRegExStr[2] != None:
				self.__rawPatterns += '(?:' + patternRegExStr[2] + ')'

		self.__compiledPatterns = None
	#



	def registerTypeParsingDelegate(self, tokenTypeMain, tokenTypeExtra, delegate):
		assert isinstance(tokenTypeMain, str)
		assert isinstance(tokenTypeExtra, (type(None), str))
		assert callable(delegate)

		if tokenTypeExtra is None:
			tokenType = tokenTypeMain
		else:
			tokenType = tokenTypeMain + "_" + tokenTypeExtra
		self.__typeParsingDelegates[tokenType] = delegate
	#



	#
	# @return		Token[] tokens			Returns token objects according to the pattern defined during initialization.
	#
	def tokenize(self, text, bEmitWhiteSpaces = False, bEmitNewLines = False):
		if not self.isCompiled:
			self.compile()

		lineNo = 1
		line_start = 0
		for mo in self.__compiledPatterns.finditer(text):
			tokenType = mo.lastgroup
			value = mo.group(tokenType)
			columnNo = mo.start() - line_start + 1
			if tokenType == 'NEWLINE':
				if bEmitNewLines:
					yield Token(tokenType, value, lineNo, columnNo)
				line_start = mo.end()
				lineNo += 1
			elif tokenType == 'SPACE':
				if bEmitWhiteSpaces:
					yield Token(tokenType, value, lineNo, columnNo)
			elif tokenType == 'ERROR':
				raise RuntimeError("Tokenization error encountered at " + str(lineNo) + ":" + str(columnNo) + "!")
			else:
				d = self.__typeParsingDelegates.get(tokenType, None)
				if d != None:
					value = d(value)
				pos = tokenType.find("_")
				# tokenTypeExtra = None
				#if pos >= 0:
					#tokenTypeExtra = tokenType[pos + 1:]
					#tokenType = tokenType[0:pos]
				# s = "" if tokenTypeExtra == None else tokenTypeExtra
				# print("---- " + tokenTypeMain + ":" + s + " ---- \"" + tokenText + "\"")
				if pos >= 0:
					tokenType = tokenType[0:pos]
				yield Token(tokenType, value, lineNo, columnNo)
				if tokenType == 'newline':
					line_start = mo.end()
					lineNo += 1
	#



	#
	# Override this method to perform correct value parsing. This method is called for every
	#
	def _parseValue(self, tokenTypeMain, tokenTypeExtra, tokenText):
		return tokenText
	#

#














import sys
from typing import Union

from jk_version import Version




class PythonModuleInfo(object):

	def __init__(self, m):
		try:
			self.location = m.__file__
			if self.location.endswith("/__init__.py"):
				self.location = self.location[:-len("/__init__.py")]
		except:
			self.location = "???"
		try:
			self.version = m.__version__
			self.__v = Version(m.__version__)
		except:
			self.version = "???"
			self.__v = None
		self.name = m.__name__
	#

	def __str__(self):
		return self.name + " : " + self.version
	#

	def __repr__(self):
		return "<PythonModuleInfo: " + self.name + ", " + self.version + ", " + self.location + ">"
	#

	def isAtLeastVersion(self, expectedMinimumVersion:Union[str,Version]):
		if isinstance(expectedMinimumVersion, str):
			expectedMinimumVersion = Version(expectedMinimumVersion)
		else:
			assert isinstance(expectedMinimumVersion, Version)

		if self.__v is None:
			return False
		else:
			return expectedMinimumVersion >= self.__v
	#

#


 
class PythonModules(object):

	def __init__(self):
		self.__moduleInfos = self.__listLoadedModules()
	#

	def dump(self, prefix:str = "", printFunction = None):
		if printFunction is None:
			printFunction = print
		else:
			assert callable(printFunction)

		for moduleInfo in self.__moduleInfos:
			printFunction(prefix + str(moduleInfo))
	#

	def moduleInfos(self):
		return self.__moduleInfos
	#

	def getModuleInfo(self, moduleName):
		for m in self.__moduleInfos:
			if m.name == moduleName:
				return m
		return None
	#

	def ensureRequirementsAreMatched(self, moduleNameVersionMap:dict):
		for moduleName, requiredVersion in moduleNameVersionMap.items():
			m = self.getModuleInfo(moduleName)
			if m is None:
				raise Exception("No such module: " + moduleName)

			if requiredVersion:
				if m.isAtLeastVersion(requiredVersion):
					# okay!
					pass
				else:
					raise Exception("Version of module " + m.name + " is " + m.version + " but version " + str(requiredVersion) + " is required!")
	#

	def __listLoadedModules(self) -> tuple:
		moduleNames = set()
		for m in sys.modules:
			p = m.find(".")
			if p > 0:
				moduleNames.add(m[:p])
		moduleNames = sorted(moduleNames)
		#modulenames = set(asys.modules) & set(globals())
		allmodules = [sys.modules[name] for name in moduleNames]
		return tuple([ PythonModuleInfo(m) for m in allmodules ])
	#

#








class MutableString(object):

	def __init__(self, text = "", width = None):
		self.__text = text
		if width != None:
			self.ensureWidth(width)
	#

	def ensureWidth(self, w):
		while len(self.__text) < w:
			self.__text += " "
	#

	def __len__(self):
		return len(self.__text)
	#

	def __str__(self):
		return self.__text
	#

	def __repr__(self):
		return self.__text
	#

	def __eq__(self, other):
		if isinstance(other, str):
			return self.__text == other
		elif isinstance(other, MutableString):
			return self.__text == other.__text
		else:
			return False
	#

	def __ne__(self, other):
		if isinstance(other, str):
			return self.__text != other
		elif isinstance(other, MutableString):
			return self.__text != other.__text
		else:
			return True
	#

	def setChar(self, x, char):
		self.ensureWidth(x + 1)
		self.__text = self.__text[0:x] + char[0] + self.__text[x+1:]
	#

	def setText(self, x, textString):
		self.ensureWidth(x + len(textString))
		self.__text = self.__text[0:x] + textString + self.__text[x+len(textString):]
	#

#















from ..Stack import Stack
from ..TypedValue import TypedValue
from .AbstractTokenPattern import AbstractTokenPattern





#
# A token pattern repeat pattern. The specified patteren must match at least once.
#
# @param		AbstractTokenPattern tokenPattern			(required) A token pattern that might be repeated multiple times.
#
class TokenPatternRepeat(AbstractTokenPattern):

	#
	# @param		AbstractTokenPattern tokenPattern			(required) A token pattern to match one or more times.
	#
	def __init__(self, tokenPattern:AbstractTokenPattern):
		super().__init__()

		assert isinstance(tokenPattern, AbstractTokenPattern)
		self.__tokenPattern = tokenPattern
	#

	def _tryMatch(self, tokens, pos, stack):
		assert isinstance(tokens, list)
		assert isinstance(pos, int)
		assert isinstance(stack, Stack)

		orgPos = pos
		nCountLoops = 0
		while True:
			(bResult, n) = self.__tokenPattern._tryMatch(tokens, pos, stack)
			assert isinstance(bResult, bool)
			assert isinstance(n, int)

			if not bResult:
				break
			pos += n
			nCountLoops += 1
		if nCountLoops > 0:
			self._addTagsToStack(stack)
			return (True, pos - orgPos)
		else:
			return (False, 0)
	#

	#
	# Derive a new token pattern from this object, modifying some attributes in the process.
	#
	def derive(self, assignToVar:str = None, assignToVarTyped:str = None, bVarIsArray:bool = False):
		assert isinstance(assignToVar, (type(None), str))
		assert isinstance(assignToVarTyped, (type(None), str))
		assert isinstance(bVarIsArray, bool)

		return TokenPatternRepeat(self.__tokenPattern.derive(assignToVar, assignToVarTyped, bVarIsArray))
	#

#











import typing

from ..ImplementationError import ImplementationError






class TypedList(list):

	################################################################################################################################
	## Constructors
	################################################################################################################################

	def __init__(self, items = None, dataType:type = None, checkDataTypeCallback = None, keyFunc = None):
		tempList = tuple(items) if (items is not None) else None

		super().__init__()

		if (dataType is None) and (checkDataTypeCallback is None):
			raise Exception("On construction either specify a data type or a callback that verifies the data type!")

		self.__nCompareWith = -1

		if dataType is not None:
			assert isinstance(dataType, type)
			self.__nCompareWith = 1
		self.__dataType = dataType
		self.__checkDataType = self.__default_checkValue

		if checkDataTypeCallback is not None:
			assert callable(checkDataTypeCallback)
			self.__nCompareWith = 2
			self.__checkDataType = checkDataTypeCallback

		if keyFunc is not None:
			assert callable(keyFunc)
		self.__keyFunc = keyFunc

		if tempList is not None:
			for item in tempList:
				self.__checkDataType(item)
			super().extend(tempList)
	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	@property
	def dataType(self) -> type:
		return self.__dataType
	#

	################################################################################################################################
	## Helper Methods
	################################################################################################################################

	def __default_checkValue(self, item) -> bool:
		return isinstance(item, self.__dataType)
	#

	def __checkDataTypeE(self, item):
		if not self.__checkDataType(item):
			raise TypeError(repr(item) + " (" + type(item).__name__ + ")")
	#

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	def append(self, item):
		self.__checkDataTypeE(item)
		super().append(item)
	#

	def extend(self, iterable):
		tempList = tuple(iterable)
		for item in tempList:
			self.__checkDataTypeE(item)
		super().extend(tempList)
	#

	def insert(self, index:int, item):
		self.__checkDataTypeE(item)
		super().insert(index, item)
	#

	def sort(self, reverse:bool = False):
		super().sort(key=self.__keyFunc, reverse=reverse)
	#

	def copy(self):
		return TypedList(
			items = self,
			dataType = self.__dataType,
			checkDataTypeCallback = self.__checkDataType,
			keyFunc = self.__keyFunc,
		)
	#

	def clone(self):
		return TypedList(
			items = self,
			dataType = self.__dataType,
			checkDataTypeCallback = self.__checkDataType,
			keyFunc = self.__keyFunc,
		)
	#

	def isCompatibleSequence(self, sequence) -> bool:
		for item in sequence:
			if not self.__checkDataType(item):
				return False
		return True
	#

	def isCompatibleSequenceE(self, sequence) -> bool:
		for item in sequence:
			if not self.__checkDataType(item):
				raise TypeError("Sequence contains values of incompatible type!")
		return True
	#

	def isCompatibleCollection(self, otherList) -> bool:
		if isinstance(otherList, TypedList):
			if self.__nCompareWith != otherList.__nCompareWith:
				return False

			if self.__nCompareWith == 1:
				if self.__dataType != otherList.__dataType:
					return False
			elif self.__nCompareWith == 2:
				if self.__checkDataType != otherList.__checkDataType:
					return False
			else:
				raise ImplementationError()

			return True
		else:
			return False
	#

	def isCompatibleCollectionE(self, otherList):
		if isinstance(otherList, TypedList):
			if self.__nCompareWith != otherList.__nCompareWith:
				raise TypeError("Other list is of incompatible type!")

			if self.__nCompareWith == 1:
				if self.__dataType != otherList.__dataType:
					raise TypeError("Other list is of incompatible type!")
			elif self.__nCompareWith == 2:
				if self.__checkDataType != otherList.__checkDataType:
					raise TypeError("Other list is of incompatible type!")
			else:
				raise TypeError("Other list is of incompatible type!")

			return

		raise TypeError("Other list is of incompatible type!")
	#

	def __add__(self, otherList):
		if self.isCompatibleCollection(otherList):
			ret = self.clone()
			ret.extend(otherList)
			return ret

		if isinstance(otherList, (tuple,list)):
			if self.isCompatibleSequence(otherList):
				ret = self.clone()
				ret.extend(otherList)
				return ret
			else:
				return NotImplemented

		return NotImplemented
	#

	def __iadd__(self, otherList):
		if self.isCompatibleCollection(otherList):
			self.extend(otherList)
			return self

		if isinstance(otherList, (tuple,list)):
			if self.isCompatibleSequence(otherList):
				self.extend(otherList)
				return self
			else:
				return NotImplemented

		return NotImplemented
	#

	def __mul__(self, n):
		if not isinstance(n, int):
			return NotImplemented

		items = super().__mul__(n)
		return TypedList(
			items = items,
			dataType = self.__dataType,
			checkDataTypeCallback = self.__checkDataType,
			keyFunc = self.__keyFunc,
		)
	#

	def __rmul__(self, n):
		if not isinstance(n, int):
			return NotImplemented

		items = super().__mul__(n)
		return TypedList(
			items = items,
			dataType = self.__dataType,
			checkDataTypeCallback = self.__checkDataType,
			keyFunc = self.__keyFunc,
		)
	#

	def __imul__(self, n):
		if not isinstance(n, int):
			return NotImplemented

		items = super().__mul__(n)
		self.clear()
		super().extend(items)
	#

	def __setitem__(self, ii, item):
		assert isinstance(ii, int)
		self.__checkDataTypeE(item)
		super().__setitem__(ii, item)
	#

	"""
	All necessary methods have been implemented, except:
	* __reduce__()
	* __reduce_ex__()
	"""

#











import datetime
import re

from .D import D












def dateRange(fromDate:D, toDate:D, includeRightBorder:bool = False):
	assert isinstance(fromDate, D)
	assert isinstance(toDate, D)
	assert isinstance(includeRightBorder, bool)

	i = fromDate.clone()
	assert isinstance(i, D)
	while i < toDate:
		yield i.clone()
		i += 1
	if i == toDate:
		if includeRightBorder:
			yield i.clone()
#

def dateRangeI(fromDate:D, toDate:D, includeRightBorder:bool = False):
	assert isinstance(fromDate, D)
	assert isinstance(toDate, D)
	assert isinstance(includeRightBorder, bool)

	n = 0
	i = fromDate.clone()
	assert isinstance(i, D)
	while i < toDate:
		yield n, i.clone()
		i += 1
		n += 1
	if i == toDate:
		if includeRightBorder:
			yield n, i.clone()
#













import os
import time
import threading





def run(delaySeconds:float, func, *args, **kwargs):
	assert isinstance(delaySeconds, (int,float))
	assert callable(func)

	def _threadFunc():
		time.sleep(delaySeconds)
		func(*args, **kwargs)
	#

	t = threading.Thread(target=_threadFunc)
	t.start()
#










__HEXCODE = "0123456789abcdef"




def byteToHex(someValue:int) -> str:
	assert isinstance(someValue, int)
	someValue = someValue & 255
	return __HEXCODE[int(someValue / 16)] + __HEXCODE[someValue % 16]

def byteArrayToHexStr(someByteArray) -> str:
	assert isinstance(someByteArray, (bytes, bytearray))
	ret = ""
	for someValue in someByteArray:
		assert isinstance(someValue, int)
		someValue = someValue & 255
		ret += __HEXCODE[int(someValue / 16)] + __HEXCODE[someValue % 16]
	return ret

def hexStrToByteArray(someHexArray:str) -> bytearray:
	if (len(someHexArray) % 2) != 0:
		raise Exception("Not a valid hex string!")
	someHexArray = someHexArray.lower()
	dataArray = bytearray()
	for offset in range(0, len(someHexArray), 2):
		charA = someHexArray[offset]
		charB = someHexArray[offset + 1]
		pA = __HEXCODE.find(charA)
		pB = __HEXCODE.find(charB)
		if (pA < 0) or (pB < 0):
			raise Exception("Not a valid hex string!")
		dataArray.append(pA * 16 + pB)
	return dataArray

def hexToByte(someHexString:str, offset:int) -> int:
	someHexString = someHexString.lower()
	charA = someHexString[offset]
	charB = someHexString[offset + 1]
	pA = __HEXCODE.find(charA)
	pB = __HEXCODE.find(charB)
	if (pA < 0) or (pB < 0):
		raise Exception("Not a valid hex string!")
	return pA * 16 + pB







import re



def _detidy_cb(m):
	if m.group(2): return m.group(2)
	if m.group(3): return m.group(3)
	return ""
#

#
# Compacts a verbose regular expression
#
def compactVerboseRegEx(retext:str) -> str:
	assert isinstance(retext, str)

	decomment = re.compile(r"""(?#!py/mx decomment Rev:20160225_1800)
		# Discard whitespace, comments and the escapes of escaped spaces and hashes.
		( (?: \s+                  # Either g1of3 $1: Stuff to discard (3 types). Either ws,
			| \#.*                   # or comments,
			| \\(?=[\r\n]|$)         # or lone escape at EOL/EOS.
			)+                       # End one or more from 3 discardables.
		)                          # End $1: Stuff to discard.
		| ( [^\[(\s#\\]+             # Or g2of3 $2: Stuff to keep. Either non-[(\s# \\.
		| \\[^# Q\r\n]             # Or escaped-anything-but: hash, space, Q or EOL.
		| \(                       # Or an open parentheses, optionally
			(?:\?\#[^)]*(?:\)|$))?   # starting a (?# Comment group).
		| \[\^?\]? [^\[\]\\]*      # Or Character class. Allow unescaped ] if first char.
			(?:\\[^Q][^\[\]\\]*)*    # {normal*} Zero or more non-[], non-escaped-Q.
			(?:                      # Begin unrolling loop {((special1|2) normal*)*}.
			(?: \[(?::\^?\w+:\])?  # Either special1: "[", optional [:POSIX:] char class.
			| \\Q       [^\\]*     # Or special2: \Q..\E literal text. Begin with \Q.
				(?:\\(?!E)[^\\]*)*   # \Q..\E contents - everything up to \E.
				(?:\\E|$)            # \Q..\E literal text ends with \E or EOL.
			)        [^\[\]\\]*    # End special: One of 2 alternatives {(special1|2)}.
			(?:\\[^Q][^\[\]\\]*)*  # More {normal*} Zero or more non-[], non-escaped-Q.
			)* (?:\]|\\?$)           # End character class with ']' or EOL (or \\EOL).
		| \\Q       [^\\]*         # Or \Q..\E literal text start delimiter.
			(?:\\(?!E)[^\\]*)*       # \Q..\E contents - everything up to \E.
			(?:\\E|$)                # \Q..\E literal text ends with \E or EOL.
		)                          # End $2: Stuff to keep.
		| \\([# ])                   # Or g3of3 $6: Escaped-[hash|space], discard the escape.
		""", re.VERBOSE | re.MULTILINE)
	return re.sub(decomment, _detidy_cb, retext)
#














import os
import sys

_has_netifaces = False
try:
	import netifaces
	_has_netifaces = True
except:
	pass





#
# Retrieve all MAC addresses from all active network adapters (excluding local host)
#
def getMACs():
	if _has_netifaces:
		ret = []

		for i in netifaces.interfaces():
			addrs = netifaces.ifaddresses(i)
			try:
				if_mac = addrs[netifaces.AF_LINK][0]['addr']
				if_ip = addrs[netifaces.AF_INET][0]['addr']
				if if_ip == "127.0.0.1":
					continue
				ret.append(if_mac)
			except IndexError as e:
				pass
			except KeyError as e:
				pass

		return ret

	else:
		raise Exception("Python module netiface module not available!")
#









import os
import sys
import socket
import time
import json
import codecs
import re

from .dtutils import *



class Cache(object):

	def __init__(self, cacheDurationSeconds):
		self.__cacheRecords = {}
		self.__cacheDurationMS = cacheDurationSeconds * 1000

	def put(self, key, value):
		self.__cacheRecords[key] = (dtutils.nowInSecondsSinceEpoch(), value)

	def get(self, key, touch = False):
		r = self.__cacheRecords.get(key, None)
		if r is None:
			return None
		(t, value) = r
		now = dtutils.nowInSecondsSinceEpoch()
		if t + self.__cacheDurationMS < now:
			del self.__cacheRecords[key]
			return None
		if touch:
			self.__cacheRecords[key] = (now, value)
		return value

	def clear(self):
		self.__cacheRecords.clear()

	def cleanup(self):
		keysToDelete = []
		now = dtutils.nowInSecondsSinceEpoch()
		for (key, (t, value)) in self.__cacheRecords.items():
			if t + self.__cacheDurationMS < now:
				keysToDelete.append(key)
		for key in keysToDelete:
			del self.__cacheRecords[key]


















import sys
import os
import inspect







def getMainModule():
	# based on: https://stackoverflow.com/questions/990422/how-to-get-a-reference-to-current-modules-attributes-in-python

	# search for first module in the stack
	stack_frame = inspect.currentframe()
	while stack_frame:
		#print('***', stack_frame.f_code.co_name, stack_frame.f_code.co_filename, stack_frame.f_lineno)
		if stack_frame.f_code.co_name == '<module>':
			if stack_frame.f_code.co_filename != '<stdin>':
				caller_module = inspect.getmodule(stack_frame)
			else:
				# piped or interactive import
				caller_module = sys.modules['__main__']
			if not caller_module is None:
				#... do something here ...
				#print("X")
				#for k, v in stack_frame.f_globals.items():
				#	print("-- " + k + ":", v)
				return sys.modules[stack_frame.f_globals["__name__"]]
			break
		stack_frame = stack_frame.f_back
#

def getMainModuleFilePath():
	mainModule = getMainModule()
	assert mainModule
	return os.path.realpath(mainModule.__file__)
#

def getMainModuleDirPath():
	mainModule = getMainModule()
	assert mainModule
	return os.path.dirname(os.path.realpath(mainModule.__file__))
#





















_GREGORIAN_CALENDAR_YEAR_IN_DAYS = 365.2425
_GREGORIAN_CALENDAR_YEAR_IN_SECONDS = 365.2425 * 24*60*60
_GREGORIAN_CALENDAR_AVG_DAYS_PER_MONTH = _GREGORIAN_CALENDAR_YEAR_IN_DAYS / 12
_GREGORIAN_CALENDAR_AVG_SECONDS_PER_MONTH = _GREGORIAN_CALENDAR_YEAR_IN_DAYS * 24*60*60 / 12
_GREGORIAN_CALENDAR_SECONDS_PER_WEEK = 24*60*60*7
_GREGORIAN_CALENDAR_SECONDS_PER_DAY = 24*60*60
_GREGORIAN_CALENDAR_SECONDS_PER_HOUR = 60*60
_GREGORIAN_CALENDAR_SECONDS_PER_MINUTE = 60










#
# Splits the specified number of seconds into years, months, weeks, days, hours, minutes and seconds.
#
# As long as the time remains below the amount of one month all return values can be calculated absolutely accurately. In that case
# <c>(0, 0, weeks, days, hours, minutes, seconds)</c> is returned.
#
# If the time is greater or equal to the duration of one month this can no longer be calculated accurately, as the length of the months differ. In that case 
# <c>(years, months, weeks, -1, -1, -1, -1)</c> is returned.
# As this function is intended to calculate values for displaying time durations, a better accuracy doesn't make any sense anyway.
#
def secondsToYMWDHMS(fSeconds:float) -> tuple:
	assert isinstance(fSeconds, (float,int))
	assert fSeconds >= 0

	if fSeconds >= _GREGORIAN_CALENDAR_YEAR_IN_SECONDS:
		years = int(fSeconds / _GREGORIAN_CALENDAR_YEAR_IN_SECONDS)
		fSeconds = fSeconds - years * _GREGORIAN_CALENDAR_YEAR_IN_SECONDS
	else:
		years = 0

	if fSeconds >= _GREGORIAN_CALENDAR_AVG_SECONDS_PER_MONTH:
		months = int(fSeconds / _GREGORIAN_CALENDAR_AVG_SECONDS_PER_MONTH)
		fSeconds = fSeconds - months * _GREGORIAN_CALENDAR_AVG_SECONDS_PER_MONTH
	else:
		months = 0

	if fSeconds >= _GREGORIAN_CALENDAR_SECONDS_PER_WEEK:
		weeks = int(fSeconds / _GREGORIAN_CALENDAR_SECONDS_PER_WEEK)
		fSeconds = fSeconds - weeks * _GREGORIAN_CALENDAR_SECONDS_PER_WEEK
	else:
		weeks = 0

	if (years > 0) or (months > 0):
		return (years, months, weeks, -1, -1, -1, -1)

	if fSeconds >= _GREGORIAN_CALENDAR_SECONDS_PER_DAY:
		days = int(fSeconds / _GREGORIAN_CALENDAR_SECONDS_PER_DAY)
		fSeconds = fSeconds - days * _GREGORIAN_CALENDAR_SECONDS_PER_DAY
	else:
		days = 0

	if fSeconds >= _GREGORIAN_CALENDAR_SECONDS_PER_HOUR:
		hours = int(fSeconds / _GREGORIAN_CALENDAR_SECONDS_PER_HOUR)
		fSeconds = fSeconds - hours * _GREGORIAN_CALENDAR_SECONDS_PER_HOUR
	else:
		hours = 0

	if fSeconds >= _GREGORIAN_CALENDAR_SECONDS_PER_MINUTE:
		minutes = int(fSeconds / _GREGORIAN_CALENDAR_SECONDS_PER_MINUTE)
		fSeconds = fSeconds - minutes * _GREGORIAN_CALENDAR_SECONDS_PER_MINUTE
	else:
		minutes = 0

	seconds = int(fSeconds)

	return (years, months, weeks, days, hours, minutes, seconds)
#




_NAMES_EN = {
	"second": ( "1 second", "{} seconds" ),
	"minute": ( "1 minute", "{} minutes" ),
	"hour": ( "1 hour", "{} hours" ),
	"day": ( "1 day", "{} days" ),
	"week": ( "1 week", "{} weeks" ),
	"month": ( "1 month", "{} months" ),
	"year": ( "1 year", "{} years" ),
}



def secondsToHRStr(fSeconds:float) -> str:
	y, m, w, d, h, m2, s = secondsToYMWDHMS(fSeconds)

	bNoWeeks = False
	bNoHours = False
	bNoMinutes = False
	bNoSeconds = False

	if y <= 0:
		sYears = ""
	else:
		_names = _NAMES_EN["year"]
		sYears = _names[0] if y == 1 else _names[1].format(y)
		bNoWeeks = True
		bNoHours = True
		bNoMinutes = True
		bNoSeconds = True

	if m <= 0:
		sMonths = ""
	else:
		_names = _NAMES_EN["month"]
		sMonths = _names[0] if m == 1 else _names[1].format(m)
		bNoHours = True
		bNoMinutes = True
		bNoSeconds = True

	if bNoWeeks or (w <= 0):
		sWeeks = ""
	else:
		_names = _NAMES_EN["week"]
		sWeeks = _names[0] if w == 1 else _names[1].format(w)
		bNoHours = True
		bNoMinutes = True
		bNoSeconds = True

	if d <= 0:
		sDays = ""
	else:
		_names = _NAMES_EN["day"]
		sDays = _names[0] if d == 1 else _names[1].format(d)
		bNoMinutes = True
		bNoSeconds = True

	if bNoHours or (h <= 0):
		sHours = ""
	else:
		_names = _NAMES_EN["hour"]
		sHours = _names[0] if h == 1 else _names[1].format(h)
		bNoSeconds = True

	if bNoMinutes or (m2 <= 0):
		sMinutes = ""
	else:
		_names = _NAMES_EN["minute"]
		sMinutes = _names[0] if m2 == 1 else _names[1].format(m2)

	if bNoSeconds or (s <= 0):
		sSeconds = ""
	else:
		_names = _NAMES_EN["second"]
		sSeconds = _names[0] if s == 1 else _names[1].format(s)

	col = [
		sYears,
		sMonths,
		sWeeks,
		sDays,
		sHours,
		sMinutes,
		sSeconds,
	]
	ret = ", ".join([ item for item in col if item ])
	if ret:
		return ret
	else:
		return "now"
#










import os
import string
import random














class RandomStateID(object):

	def __init__(self):
		self.__randomStateID = self.__randomString()
	#

	def __randomString(self, stringLength:int = 32) -> str:
		letters = string.ascii_letters + string.digits
		return "".join(random.choice(letters) for i in range(stringLength))
	#

	def touch(self) -> str:
		self.__randomStateID = self.__randomString()
		return self.__randomStateID
	#

	@property
	def randomStateID(self) -> str:
		return self.__randomStateID
	#

	def __str__(self):
		return self.__randomStateID
	#

#

























import time
import threading
from typing import Union

import sortedcontainers





class AsyncActivityRecord(object):

	def __init__(self, identifier, timeToRun:float, theCallable:callable, callArguments:Union[tuple,list], autoRescheduleDelay:Union[int,float] = 0):
		self.identifier = identifier
		self.timeToRun = timeToRun
		self.theCallable = theCallable
		self.callArguments = callArguments
		self.autoRescheduleDelay = autoRescheduleDelay
	#

#



def _getARKey(item):
	return item.timeToRun
#



class AsyncRunner(object):

	def __init__(self, debugLogPrintFunction:callable = None):
		# list of all activities
		self.__activitiesByTime = sortedcontainers.SortedListWithKey(key=_getARKey)
		# list of activities that later can be identified again
		self.__activitiesByID = dict()
		self.__debugLogPrintFunction = debugLogPrintFunction
		self.__lock = threading.Lock()
		self.__evt = threading.Event()
		self.__thread = threading.Thread(target=self.__run, daemon=True)
		self.__bKeepRunning = True
		self.__bRunning = True
		self.__bIsStarted = False
	#

	def start(self):
		if self.__bIsStarted:
			if self.__bRunning:
				return True
			else:
				return False
		else:
			self.__bIsStarted = True
			self.__thread.start()
			return True
	#

	def __del__(self):
		if self.__bRunning:
			self.stop()
	#

	def __run(self):
		if self.__debugLogPrintFunction:
			self.__debugLogPrintFunction("[AsyncRunner] Thread started.")

		timeToWait = None
		while self.__bKeepRunning:
			if self.__debugLogPrintFunction:
				self.__debugLogPrintFunction("[AsyncRunner] Queue length is " + str(len(self.__activitiesByTime)) + ". Waiting.")

			self.__evt.wait(timeToWait)

			tNow = time.time()
			while True:

				if self.__debugLogPrintFunction:
					self.__debugLogPrintFunction("[AsyncRunner] Queue length is " + str(len(self.__activitiesByTime)) + ". Checking.")

				# get top most activity
				self.__lock.acquire()
				if len(self.__activitiesByTime) > 0:
					# activity found; remove it;
					topMostActivity = self.__activitiesByTime[0]
					self.__activitiesByTime.remove(topMostActivity)
					if topMostActivity.identifier != None:
						del self.__activitiesByID[topMostActivity.identifier]
					if self.__debugLogPrintFunction:
						self.__debugLogPrintFunction("[AsyncRunner] Retrieving from queue: " + str(topMostActivity))
				else:
					topMostActivity = None
				self.__lock.release()

				if topMostActivity:
					# we have another activity
					if topMostActivity.timeToRun <= tNow:
						# activity is scheduled to run now!
						if self.__debugLogPrintFunction:
							self.__debugLogPrintFunction("[AsyncRunner] Activitiy to run: " + str(topMostActivity))

						if topMostActivity.autoRescheduleDelay > 0:
							# reschedule the activity

							if self.__debugLogPrintFunction:
								self.__debugLogPrintFunction("[AsyncRunner] Reinserting to run in " + str(topMostActivity.autoRescheduleDelay) + " seconds: " + str(topMostActivity))
							topMostActivity.timeToRun = tNow + topMostActivity.autoRescheduleDelay

							self.__lock.acquire()
							self.__activitiesByTime.add(topMostActivity)
							if topMostActivity.identifier != None:
								self.__activitiesByID[topMostActivity.identifier] = topMostActivity
							self.__lock.release()

						# now run the activity
						if self.__debugLogPrintFunction:
							self.__debugLogPrintFunction("[AsyncRunner] Now running: " + str(topMostActivity))
						topMostActivity.theCallable(topMostActivity.callArguments)
					else:
						# wait
						timeToWait = topMostActivity.timeToRun - tNow

						if self.__debugLogPrintFunction:
							self.__debugLogPrintFunction("[AsyncRunner] Reinserting to run in " + str(timeToWait) + " seconds: " + str(topMostActivity))

						self.__lock.acquire()
						self.__activitiesByTime.add(topMostActivity)
						if topMostActivity.identifier != None:
							self.__activitiesByID[topMostActivity.identifier] = topMostActivity
						self.__lock.release()

						break
				else:
					# no more activities waiting
					timeToWait = None
					if self.__debugLogPrintFunction:
						self.__debugLogPrintFunction("[AsyncRunner] Queue is empty.")
					break

		if self.__debugLogPrintFunction:
			self.__debugLogPrintFunction("[AsyncRunner] Thread terminated.")

		self.__bRunning = False
	#

	def stop(self):
		self.__bKeepRunning = False
		self.__evt.set()
		self.__evt.clear()
	#

	def rescheduleCallable(self, theCallable:callable, callArguments:Union[tuple,list], delaySeconds:Union[int,float],
		identifier:Union[str,int,None] = None, autoRescheduleDelay:Union[int,float] = 0):

		assert callable(theCallable)
		assert isinstance(delaySeconds, (int, float))
		assert delaySeconds >= 0
		assert isinstance(autoRescheduleDelay, (int, float))

		if identifier is None:
			identifier = id(theCallable)
		else:
			isinstance(identifier, (str, int))

		timeToRun = time.time() + delaySeconds

		if self.__debugLogPrintFunction:
			self.__debugLogPrintFunction("[AsyncRunner] (re)scheduling activity to run in " + str(delaySeconds) + " seconds: " + str(theCallable) + ", " + repr(callArguments))

		self.__lock.acquire()
		existingActivity = self.__activitiesByID.get(identifier, None)
		if existingActivity:
			self.__activitiesByTime.remove(existingActivity)
			existingActivity.timeToRun = timeToRun
			existingActivity.theCallable = theCallable
			existingActivity.callArguments = callArguments
			self.__activitiesByTime.add(existingActivity)
		else:
			newActivity = AsyncActivityRecord(identifier, timeToRun, theCallable, callArguments, autoRescheduleDelay)
			self.__activitiesByTime.add(newActivity)
			self.__activitiesByID[identifier] = newActivity
		self.__lock.release()

		self.__evt.set()
		self.__evt.clear()
	#

	def removeScheduledCallable(self, identifier:Union[str,int]):
		isinstance(identifier, (str, int))

		self.__lock.acquire()
		existingActivity = self.__activitiesByID.get(identifier, None)
		if existingActivity:
			self.__activitiesByTime.remove(existingActivity)
			del self.__activitiesByID[identifier]

			if self.__debugLogPrintFunction:
				self.__debugLogPrintFunction("[AsyncRunner] Removed: " + str(existingActivity))
		self.__lock.release()
	#

	def scheduleCallable(self, theCallable:callable, callArguments:Union[tuple,list], delaySeconds:Union[int,float] = 0,
		autoRescheduleDelay:Union[int,float] = 0):

		assert callable(theCallable)
		assert isinstance(delaySeconds, (int, float))
		assert delaySeconds >= 0
		assert isinstance(autoRescheduleDelay, (int, float))

		timeToRun = time.time() + delaySeconds

		if self.__debugLogPrintFunction:
			self.__debugLogPrintFunction("[AsyncRunner] scheduling activity to run in " + str(delaySeconds) + " seconds: " + str(theCallable) + ", " + repr(callArguments))

		newActivity = AsyncActivityRecord(None, timeToRun, theCallable, callArguments, autoRescheduleDelay)
		self.__lock.acquire()
		self.__activitiesByTime.add(newActivity)
		self.__lock.release()

		self.__evt.set()
		self.__evt.clear()
	#

#










